"""Collection of functions for the manipulation of time series."""

import warnings

import cltoolbox
import numpy as np
from cltoolbox.rst_text_formatter import RSTHelpFormatter
from pydantic import PositiveInt, conint, validate_arguments
from toolbox_utils import tsutils

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


warnings.filterwarnings("ignore")


@validate_arguments
def _boolrelextrema(data, comparator, axis=0, order: PositiveInt = 1, mode="clip"):
    """Calculate the relative extrema of `data`.

    Relative extrema are calculated by finding locations where
    comparator(data[n],data[n+1:n+order+1]) = True.

    Parameters
    ----------
    data: ndarray
    comparator: function
        function to use to compare two data points.  Should take 2 numbers as
        arguments
    axis: int, optional
        axis over which to select from `data`
    order: int, optional
        How many points on each side to require a `comparator`(n,n+x) = True.
    mode: string, optional
        How the edges of the vector are treated.  'wrap' (wrap around) or
        'clip' (treat overflow as the same as the last (or first) element).
        Default 'clip'. See numpy.take

    Returns
    -------
    extrema: ndarray
        Indices of the extrema, as boolean array of same shape as data. True
        for an extrema, False else.

    See Also
    --------
    argrelmax, argrelmin

    Examples
    --------
    >>> testdata = np.array([1,2,3,2,1])
    >>> _boolrelextrema(testdata, np.greater, axis=0).tolist()
    [False, False, True, False, False]

    """
    datalen = data.shape[axis]
    locs = np.arange(0, datalen)

    results = np.ones(data.shape, dtype=bool)
    main = data.take(locs)
    for shift in range(1, order + 1):
        plus = np.take(data, locs + shift, axis=axis, mode=mode)
        results &= comparator(main, plus)
        minus = np.take(data, locs - shift, axis=axis, mode=mode)
        results &= comparator(main, minus)
        if ~results.any():
            return results
    return results


def _argrel(data, axis=0, window=1):
    """Private function to find relative min and max of data."""
    tmpmin = _argrelmin(data, axis=axis, order=window)
    tmpmax = _argrelmax(data, axis=axis, order=window)
    return (
        list(zip(tmpmax[0], data[tmpmax[0]])),
        list(zip(tmpmin[0], data[tmpmin[0]])),
    )


def _argrelmin(data, axis=0, order=1, mode="clip"):
    """Calculate the relative minima of `data`.

    See Also
    --------
    argrelextrema, argrelmax

    """
    return _argrelextrema(data, np.less, axis, order, mode)


def _argrelmax(data, axis=0, order=1, mode="clip"):
    """Calculate the relative maxima of `data`.

    See Also
    --------
    argrelextrema, argrelmin

    """
    return _argrelextrema(data, np.greater, axis, order, mode)


def _argrelextrema(data, comparator, axis=0, order=1, mode="clip"):
    """Calculate the relative extrema of `data`.

    Returns
    -------
    extrema: ndarray
        Indices of the extrema, as an array
        of integers (same format as argmin, argmax

    See Also
    --------
    argrelmin, argrelmax

    """
    results = _boolrelextrema(data, comparator, axis, order, mode)
    if ~results.any():
        return (np.array([]),) * 2
    return np.where(results)


def _datacheck_peakdetect(x_axis, y_axis):
    """Check x and y axis, creating an x data_set if necessary."""
    if x_axis is None:
        x_axis = list(range(len(y_axis)))

    if len(y_axis) != len(x_axis):
        raise ValueError(
            tsutils.error_wrapper(
                f"""
The length of y values must equal the length of x values.  Instead the
length of y values is {len(y_axis)} and the length of x values is {len(x_axis)}.
"""
            )
        )

    # needs to be a numpy array
    y_axis = np.array(y_axis)
    x_axis = np.array(x_axis)
    return x_axis, y_axis


def _peakdetect_parabola_fitter(raw_peaks, x_axis, y_axis, points):
    """Perform parabola fitting for the _peakdetect_parabola function.

    keyword arguments:
    raw_peaks -- A list of either the maximum or the minimum peaks, as given
        by the _peakdetect_zero_crossing function, with index used as x-axis
    x_axis -- A numpy list of all the x values
    y_axis -- A numpy list of all the y values
    points -- How many points around the peak should be used during curve
        fitting, must be odd.

    return -- A list giving all the peaks and the fitted waveform, format:
        [[x, y, [fitted_x, fitted_y]]]

    """
    from scipy.optimize import curve_fit

    def func(x, k, tau, m):
        return k * ((x - tau) ** 2) + m

    fitted_peaks = []
    for peak in raw_peaks:
        index = peak[0]
        x_data = x_axis[index - points // 2 : index + points // 2 + 1]
        y_data = y_axis[index - points // 2 : index + points // 2 + 1]
        # get a first approximation of tau (peak position in time)
        tau = x_axis[index]
        # get a first approximation of peak amplitude
        m = peak[1]

        # build list of approximations
        # k = -m as first approximation?
        p0 = (-m, tau, m)
        popt, _ = curve_fit(func, x_data, y_data, p0)
        # retrieve tau and m i.e x and y value of peak
        x, y = popt[1:3]

        # create a high resolution data set for the fitted waveform
        x2 = np.linspace(x_data[0], x_data[-1], points * 10)
        y2 = func(x2, *popt)

        fitted_peaks.append([x, y, [x2, y2]])

    return fitted_peaks


@validate_arguments
def _peakdetect(
    y_axis,
    x_axis=None,
    window: PositiveInt = 24,
    delta: conint(ge=0) = 0,
):
    """Private peak detection algorithm.

    Converted from/based on a MATLAB script
    at:http://billauer.co.il/peakdet.html

    function for detecting local maximas and minmias in a signal.
    Discovers peaks by searching for values which are surrounded by lower
    or larger values for maximas and minimas respectively

    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks
    x_axis -- (optional) A x-axis whose values correspond to the y_axis list
        and is used in the return to specify the position of the peaks. If
        omitted an index of the y_axis is used. (default: None)
    window -- (optional) distance to look ahead from a peak candidate to
        determine if it is the actual peak (default: 200)
        '(sample / period) / f' where '4 >= f >= 1.25' might be a good value
    delta -- (optional) this specifies a minimum difference between a peak and
        the following points, before a peak may be considered a peak. Useful
        to hinder the function from picking up false peaks towards to end of
        the signal. To work well delta should be set to delta >= RMSnoise * 5.
        (default: 0)
            delta function causes a 20% decrease in speed, when omitted
            Correctly used it can double the speed of the function

    return -- two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tuple
        of: (position, peak_value)
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do:
        x, y = zip(*tab)

    """
    max_peaks = []
    min_peaks = []
    dump = []  # Used to pop the first hit which almost always is false

    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    # store data length for later use
    length = len(y_axis)

    # maxima and minima candidates are temporarily stored in
    # mx and mn respectively
    mn, mx = np.Inf, -np.Inf

    # Only detect peak if there is 'window' amount of points after it
    for index, (x, y) in enumerate(zip(x_axis[:-window], y_axis[:-window])):
        if y > mx:
            mx = y
            mxpos = x
        if y < mn:
            mn = y
            mnpos = x

        # ####look for max####
        if (
            y < mx - delta
            and mx != np.Inf
            and y_axis[index : index + window].max() < mx
        ):
            max_peaks.append([mxpos, mx])
            dump.append(True)
            # set algorithm to only find minima now
            mx = np.Inf
            mn = np.Inf
            if index + window >= length:
                # end is within window no more peaks can be found
                break
            continue

        # ####look for min####
        if (
            y > mn + delta
            and mn != -np.Inf
            and y_axis[index : index + window].min() > mn
        ):
            min_peaks.append([mnpos, mn])
            dump.append(False)
            # set algorithm to only find maxima now
            mn = -np.Inf
            mx = -np.Inf
            if index + window >= length:
                # end is within window no more peaks can be found
                break

    # Remove the false hit on the first value of the y_axis
    try:
        if dump[0]:
            max_peaks.pop(0)
        else:
            min_peaks.pop(0)
    except IndexError:
        # no peaks were found, should the function return empty lists?
        pass

    return [max_peaks, min_peaks]


def _peakdetect_fft(y_axis, x_axis, pad_len=5):
    """Private function to calculate FFT peak detection.

    Performs a FFT calculation on the data and zero-pads the results to
    increase the time domain resolution after performing the inverse fft and
    send the data to the '_peakdetect' function for peak
    detection.

    Omitting the x_axis is forbidden as it would make the resulting x_axis
    value silly if it was returned as the index 50.234 or similar.

    Will find at least 1 less peak then the '_peakdetect_zero_crossing'
    function, but should result in a more precise value of the peak as
    resolution has been increased. Some peaks are lost in an attempt to
    minimize spectral leakage by calculating the fft between two zero
    crossings for n amount of signal periods.

    The biggest time eater in this function is the ifft and thereafter it's
    the '_peakdetect' function which takes only half the time of the ifft.
    Speed improvement could include to check if 2**n points could be used for
    fft and ifft or change the '_peakdetect' to the
    '_peakdetect_zero_crossing',
    which is maybe 10 times faster than 'peakdetct'. The pro of '_peakdetect'
    is that it results in one less lost peak. It should also be noted that the
    time used by the ifft function can change greatly depending on the input.

    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks
    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the position of the peaks.
    pad_len -- (optional) By how many times the time resolution should be
        increased by, e.g. 1 doubles the resolution. The amount is rounded up
        to the nearest 2 ** n amount (default: 5)

    return -- two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tuple
        of: (position, peak_value)
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do:
        x, y = zip(*tab)

    """
    from scipy import fft, ifft

    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    zero_indices = zero_crossings(y_axis, window=11)
    # select a n amount of periods
    last_indice = -1 - (1 - len(zero_indices) & 1)
    # Calculate the fft between the first and last zero crossing
    # this method could be ignored if the beginning and the end of the signal
    # are discarded as any errors induced from not using whole periods
    # should mainly manifest in the beginning and the end of the signal, but
    # not in the rest of the signal
    if len(zero_indices) < 2:
        fft_data = fft(y_axis)
    else:
        fft_data = fft(y_axis[zero_indices[0] : zero_indices[last_indice]])

    def padd(x, c):
        return x[: len(x) // 2] + [0] * c + x[len(x) // 2 :]

    def n(x):
        return (np.log(x) // np.log(2)).astype("i") + 1

    # pads to 2**n amount of samples
    fft_padded = padd(list(fft_data), 2 ** n(len(fft_data) * pad_len) - len(fft_data))

    # There is amplitude decrease directly proportional to the sample increase
    sf = len(fft_padded) // float(len(fft_data))
    # There might be a leakage giving the result an imaginary component
    # Return only the real component
    y_axis_ifft = ifft(fft_padded).real * sf  # (pad_len + 1)
    x_axis_ifft = np.linspace(
        x_axis[zero_indices[0]], x_axis[zero_indices[last_indice]], len(y_axis_ifft)
    )
    # get the peaks to the interpolated waveform
    max_peaks, min_peaks = _peakdetect(
        y_axis_ifft, x_axis_ifft, 500, delta=abs(np.diff(y_axis).max() * 2)
    )

    return [max_peaks, min_peaks]


def _peakdetect_parabola(y_axis, x_axis, points=9):
    """Private function for parabola peak detection.

    Function for detecting local maximas and minmias in a signal.
    Discovers peaks by fitting the model function: y = k (x - tau) ** 2 + m
    to the peaks. The amount of points used in the fitting is set by the
    points argument.

    Omitting the x_axis is forbidden as it would make the resulting x_axis
    value silly if it was returned as index 50.234 or similar.

    will find the same amount of peaks as the '_peakdetect_zero_crossing'
    function, but might result in a more precise value of the peak.

    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks
    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the position of the peaks.
    points -- (optional) How many points around the peak should be used during
        curve fitting, must be odd (default: 9)

    return -- two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a list
        of: (position, peak_value)
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do:
        x, y = zip(*max_peaks)

    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    # make the points argument odd
    points += 1 - points % 2
    # points += 1 - int(points) & 1 slower when int conversion needed

    # get raw peaks
    max_raw, min_raw = _peakdetect_zero_crossing(y_axis)

    # define output variable
    max_peaks = []
    min_peaks = []

    max_ = _peakdetect_parabola_fitter(max_raw, x_axis, y_axis, points)
    min_ = _peakdetect_parabola_fitter(min_raw, x_axis, y_axis, points)

    max_peaks = [[x[0], x[1]] for x in max_]
    min_peaks = [[x[0], x[1]] for x in min_]

    return [max_peaks, min_peaks]


def _peakdetect_sine(y_axis, x_axis, points=9, lock_frequency=False):
    """Private function for sine wave based peak detection.

    Function for detecting local maximas and minmias in a signal.
    Discovers peaks by fitting the model function:
    y = A * sin(2 * pi * f * x - tau) to the peaks. The amount of points used
    in the fitting is set by the points argument.

    Omitting the x_axis is forbidden as it would make the resulting x_axis
    value silly if it was returned as index 50.234 or similar.

    will find the same amount of peaks as the '_peakdetect_zero_crossing'
    function, but might result in a more precise value of the peak.

    The function might have some problems if the sine wave has a
    non-negligible total angle i.e. a k*x component, as this messes with the
    internal offset calculation of the peaks, might be fixed by fitting a
    k * x + m function to the peaks for offset calculation.

    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks
    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the position of the peaks.
    points -- (optional) How many points around the peak should be used during
        curve fitting, must be odd (default: 9)
    lock_frequency -- (optional) Specifies if the frequency argument of the
        model function should be locked to the value calculated from the raw
        peaks or if optimization process may tinker with it. (default: False)

    return -- two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tuple
        of: (position, peak_value)
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do:
        x, y = zip(*tab)

    """
    from scipy.optimize import curve_fit

    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    # make the points argument odd
    points += 1 - points % 2
    # points += 1 - int(points) & 1 slower when int conversion needed

    # get raw peaks
    max_raw, min_raw = _peakdetect_zero_crossing(y_axis)

    # define output variable
    max_peaks = []
    min_peaks = []

    # get global offset
    offset = np.mean([np.mean(max_raw, 0)[1], np.mean(min_raw, 0)[1]])
    # fitting a k * x + m function to the peaks might be better
    # offset_func = lambda x, k, m: k * x + m

    # calculate an approximate frequency of the signal
    hz = []
    for raw in [max_raw, min_raw]:
        if len(raw) > 1:
            peak_pos = [x_axis[index] for index in zip(*raw)[0]]
            hz.append(np.mean(np.diff(peak_pos)))
    hz = 1.0 // np.mean(hz)

    # model function
    # if cosine is used then tau could equal the x position of the peak
    # if sine were to be used then tau would be the first zero crossing
    if lock_frequency:
        # This is strange - only difference in the two "func"
        # definitions is the "hz" argument

        def func(x, a, tau):
            return a * np.sin(2 * np.pi * hz * (x - tau) + (np.pi // 2))

    else:

        def func(x, a, hz, tau):
            return a * np.sin(2 * np.pi * hz * (x - tau) + (np.pi // 2))

    # get peaks
    fitted_peaks = []
    for raw_peaks in [max_raw, min_raw]:
        peak_data = []
        for peak in raw_peaks:
            index = peak[0]
            x_data = x_axis[index - points // 2 : index + points // 2 + 1]
            y_data = y_axis[index - points // 2 : index + points // 2 + 1]
            # get a first approximation of tau (peak position in time)
            tau = x_axis[index]
            # get a first approximation of peak amplitude
            a = peak[1]

            # build list of approximations
            if lock_frequency:
                p0 = (a, tau)
            else:
                p0 = (a, hz, tau)

            # subtract offset from waveshape
            y_data -= offset
            popt, _ = curve_fit(func, x_data, y_data, p0)
            # retrieve tau and a i.e x and y value of peak
            x = popt[-1]
            y = popt[0]

            # create a high resolution data set for the fitted waveform
            x2 = np.linspace(x_data[0], x_data[-1], points * 10)
            y2 = func(x2, *popt)

            # add the offset to the results
            y += offset
            y2 += offset
            y_data += offset

            peak_data.append([x, y, [x2, y2]])

        fitted_peaks.append(peak_data)

    # structure date for output
    max_peaks = [[xval[0], xval[1]] for xval in fitted_peaks[0]]
    min_peaks = [[xval[0], xval[1]] for xval in fitted_peaks[1]]

    return [max_peaks, min_peaks]


def _peakdetect_sine_locked(y_axis, x_axis, points=9):
    """Private function for peak detection with locked sine wave.

    Convenience function for calling the '_peakdetect_sine' function with
    the lock_frequency argument as True.

    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks
    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the position of the peaks.
    points -- (optional) How many points around the peak should be used during
        curve fitting, must be odd (default: 9)

    return -- see '_peakdetect_sine'

    """
    return _peakdetect_sine(y_axis, x_axis, points, True)


def _peakdetect_zero_crossing(y_axis, x_axis=None, window=5):
    """Private function for zero crossing peak detection.

    Function for detecting local maximas and minmias in a signal.
    Discovers peaks by dividing the signal into bins and retrieving the
    maximum and minimum value of each the even and odd bins respectively.
    Division into bins is performed by smoothing the curve and finding the
    zero crossings.

    Suitable for repeatable signals, where some noise is tolerated. Executes
    faster than '_peakdetect', although this function will break if the offset
    of the signal is too large. It should also be noted that the first and
    last peak will probably not be found, as this function only can find peaks
    between the first and last zero crossing.

    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks
    x_axis -- (optional) A x-axis whose values correspond to the y_axis list
        and is used in the return to specify the position of the peaks. If
        omitted an index of the y_axis is used. (default: None)
    window -- the dimension of the smoothing window; should be an odd integer
        (default: 11)

    return -- two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tuple
        of: (position, peak_value)
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do:
        x, y = zip(*tab)

    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)

    zero_indices = zero_crossings(y_axis, window=window)
    zero_indices = np.concatenate(([0], zero_indices, [len(y_axis) - 1]))
    zero_indices = np.unique(zero_indices)

    period_lengths = np.diff(zero_indices)

    bins_y = [
        y_axis[index : index + diff]
        for index, diff in zip(zero_indices, period_lengths)
    ]
    bins_x = [
        x_axis[index : index + diff]
        for index, diff in zip(zero_indices, period_lengths)
    ]

    even_bins_y = bins_y[::2]
    odd_bins_y = bins_y[1::2]
    even_bins_x = bins_x[::2]
    odd_bins_x = bins_x[1::2]
    hi_peaks_x = []
    lo_peaks_x = []

    # check if even bin contains maxima
    if abs(even_bins_y[0].max()) > abs(even_bins_y[0].min()):
        hi_peaks = [bin.max() for bin in even_bins_y]
        lo_peaks = [bin.min() for bin in odd_bins_y]
        # get x values for peak
        hi_peaks_x.extend(
            bin_x[np.where(bin_y == peak)[0][0]]
            for bin_x, bin_y, peak in zip(even_bins_x, even_bins_y, hi_peaks)
        )

        lo_peaks_x.extend(
            bin_x[np.where(bin_y == peak)[0][0]]
            for bin_x, bin_y, peak in zip(odd_bins_x, odd_bins_y, lo_peaks)
        )

    else:
        hi_peaks = [bin.max() for bin in odd_bins_y]
        lo_peaks = [bin.min() for bin in even_bins_y]
        # get x values for peak
        hi_peaks_x.extend(
            bin_x[np.where(bin_y == peak)[0][0]]
            for bin_x, bin_y, peak in zip(odd_bins_x, odd_bins_y, hi_peaks)
        )

        lo_peaks_x.extend(
            bin_x[np.where(bin_y == peak)[0][0]]
            for bin_x, bin_y, peak in zip(even_bins_x, even_bins_y, lo_peaks)
        )

    # peaks or valley cannot be at 0
    max_peaks = [[x, y] for x, y in zip(hi_peaks_x, hi_peaks) if x != 0]
    min_peaks = [[x, y] for x, y in zip(lo_peaks_x, lo_peaks) if x != 0]

    return [max_peaks, min_peaks]


@validate_arguments
def _smooth(
    x,
    window_len=11,
    window: Literal["flat", "hanning", "hamming", "bartlett", "blackman"] = "hanning",
):
    """Smooth the data using a window of the requested size.

    This method is based on the convolution of a scaled window on the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the beginning and end parts of the output signal.

    Parameters
    ----------
    x:
        The input signal.
    window_len:
        The dimension of the smoothing window; should be an odd
        integer.
    window:
        The type of window from 'flat', 'hanning', 'hamming', 'bartlett',
        'blackman'.  The 'flat' window will produce a moving average smoothing.

    Returns
    -------
    out:
        The smoothed signal.

    Examples
    --------
        >>> t = np.linspace(-2, 2, 20)
        >>> x = np.sin(t)+np.random.randn(len(t))*0.1
        >>> y = _smooth(x)

    See Also
    --------
    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman
    numpy.convolve, scipy.signal.lfilter

    """
    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        return x

    s = np.r_[x[window_len - 1 : 0 : -1], x, x[-1:-window_len:-1]]
    if window == "flat":  # moving average
        w = np.ones(window_len, "d")
    else:
        w = eval(f"np.{window}(window_len)")

    return np.convolve((w / w.sum()), s, mode="valid")


def zero_crossings(y_axis, window=11):
    """Zero crossing peak detect.

    Algorithm to find zero crossings. Smooths the curve and finds the
    zero-crossings by looking for a sign change.

    keyword arguments:
    y_axis -- A list containing the signal over which to find zero-crossings
    window -- the dimension of the smoothing window; should be an odd integer
        (default: 11)

    return -- the index for each zero-crossing

    """
    # smooth the curve
    length = len(y_axis)
    x_axis = np.asarray(list(range(length)), int)

    ymean = y_axis.mean()
    y_axis = y_axis - ymean

    # discard tail of smoothed signal
    y_axis = _smooth(y_axis, window)[:length]
    nzero_crossings = np.where(np.diff(np.sign(y_axis)))[0]
    indices = [x_axis[index] for index in nzero_crossings]

    # check if zero-crossings are valid
    # diff = np.diff(indices)
    #    if diff.std() / diff.mean() > 0.2:
    #        print diff.std() / diff.mean()
    #        print np.diff(indices)
    #        raise(ValueError,
    #            "False zero-crossings found, indicates problem {0} or {1}".format(
    #            "with smoothing window", "problem with offset"))
    # check if any zero crossings were found
    if len(nzero_crossings) < 1:
        raise ValueError

    try:
        indices.remove(0)
    except ValueError:
        pass

    return indices
    # used this to test the fft function's sensitivity to spectral leakage
    # return indices + np.asarray(30 * np.random.randn(len(indices)), int)


# Frequency calculation#############################
#    diff = np.diff(indices)
#    time_p_period = diff.mean()
#
#    if diff.std() / time_p_period > 0.1:
#        raise ValueError,
#            "smoothing window too small, false zero-crossing found"
#
#    #return frequency
#    return 1.0 / time_p_period
##############################################################################


@cltoolbox.command("peak_detection", formatter_class=RSTHelpFormatter)
@tsutils.doc(tsutils.docstrings)
def peak_detection_cli(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    method="rel",
    extrema="peak",
    window=24,
    pad_len=5,
    points=9,
    lock_frequency=False,
    float_format="g",
    round_index=None,
    source_units=None,
    target_units=None,
    print_input="",
    tablefmt="csv",
):
    r"""Peak and valley detection.

    Parameters
    ----------
    extrema: str
        [optional, default is 'peak']

        'peak', 'valley', or 'both' to determine what should be
        returned.
    method: str
        [optional, default is 'rel']

        'rel', 'minmax', 'zero_crossing', 'parabola', 'sine' methods are
        available.  The different algorithms have different strengths
        and weaknesses.
    window: int
        [optional, default is 24]

        There will not usually be multiple peaks within the window
        number of values.  The different methods use this variable in
        different ways.  For 'rel' the window keyword specifies how many
        points on each side to require a comparator(n,n+x) = True.  For
        'minmax' the window keyword is the distance to look ahead from
        a peak candidate to determine if it is the actual peak.

        '(sample / period) / f'

        where f might be a good choice between 1.25 and 4.

        For 'zero_crossing' the window keyword is the dimension of the
        smoothing window and should be an odd integer.
    pad_len: int
        [optional, default is 5]

        Used with FFT to pad edges of time-series.
    points: int
        [optional, default is 9]

        For 'parabola' and 'sine' methods. How many points around the
        peak should be used during curve fitting, must be odd.  The
    lock_frequency
        [optional, default is False]

        For 'sine' method only.  Specifies if the frequency argument of
        the model function should be locked to the value calculated from
        the raw peaks or if optimization process may tinker with it.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${skiprows}
    ${index_type}
    ${names}
    ${clean}
    ${float_format}
    ${round_index}
    ${source_units}
    ${target_units}
    ${print_input}
    ${tablefmt}
    """
    tsutils.printiso(
        peak_detection(
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            dropna=dropna,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            clean=clean,
            method=method,
            extrema=extrema,
            window=window,
            pad_len=pad_len,
            points=points,
            lock_frequency=lock_frequency,
            round_index=round_index,
            source_units=source_units,
            target_units=target_units,
            print_input=print_input,
        ),
        float_format=float_format,
        tablefmt=tablefmt,
    )


@validate_arguments
@tsutils.copy_doc(peak_detection_cli)
def peak_detection(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    method: Literal["rel", "minmax", "zero_crossing", "parabola", "sine"] = "rel",
    extrema: Literal["peak", "valley", "both"] = "peak",
    window: PositiveInt = 24,
    pad_len: PositiveInt = 5,
    points: PositiveInt = 9,
    lock_frequency: bool = False,
    round_index=None,
    source_units=None,
    target_units=None,
    print_input="",
):
    r"""Peak and valley detection."""
    # Couldn't get fft method working correctly.  Left pieces in
    # in case want to figure it out in the future.
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )

    window = int(window)
    kwds = {}
    if method == "fft":
        func = _peakdetect_fft
        kwds["pad_len"] = int(pad_len)

    elif method == "minmax":
        func = _peakdetect
        window //= 2
        if window == 0:
            window = 1
        kwds["window"] = window
    elif method == "parabola":
        func = _peakdetect_parabola
        kwds["points"] = int(points)
    elif method == "rel":
        func = _argrel
        window /= 2
        if window == 0:
            window = 1
        kwds["window"] = window
    elif method == "sine":
        func = _peakdetect_sine
        kwds["points"] = int(points)
        kwds["lock_frequency"] = lock_frequency
    elif method == "zero_crossing":
        func = _peakdetect_zero_crossing
        if not window % 2:
            window += 1
        kwds["window"] = window
    tmptsd = tsd.copy()
    if extrema == "both":
        tmptsd.columns = [tsutils.renamer(i, "peak") for i in tsd.columns]
        atmpst = tsd.copy()
        atmpst.columns = [tsutils.renamer(i, "valley") for i in tsd.columns]
        tmptsd = tmptsd.join(atmpst, how="outer")
    else:
        tmptsd.columns = [tsutils.renamer(i, extrema) for i in tsd.columns]

    for cols in tmptsd.columns:
        if method in ["fft", "parabola", "sine"]:
            maxpeak, minpeak = func(
                tmptsd[cols].values, list(range(len(tmptsd[cols]))), **kwds
            )
        else:
            maxpeak, minpeak = func(tmptsd[cols].values, **kwds)
        if cols[-4:] == "peak":
            datavals = maxpeak
        if cols[-6:] == "valley":
            datavals = minpeak
        maxx, _ = list(zip(*datavals))
        hold = tmptsd[cols][np.array(maxx).astype("i")]
        tmptsd[cols][:] = np.nan
        tmptsd[cols][np.array(maxx).astype("i")] = hold

    tmptsd.index.name = "Datetime"
    tsd.index.name = "Datetime"
    return tsutils.return_input(print_input, tsd, tmptsd, suffix="")



import os

import xlsxwriter


def write_taylor_stats(filename, data, **kwargs):
    """
    Write statistics used in a target diagram to an Excel file.

    This function writes to an Excel file FILENAME the statistics
    used to create a Taylor diagram for each of the dictionaries
    contained in DATA. The first 2 arguments must be the inputs as
    described below followed by keyword arguments in the format of
    OPTION = VALUE.

    INPUTS:
    filename      : name for statistics Excel file
    data          : a dictionary containing the statistics
    data['sdev']  : Standard deviations (sigma)
    data['crmsd'] : Centered Root Mean Square Difference (CRMSD)
    data['ccoef'] : Correlation Coefficient (r)

    OUTPUTS:
        None.

    LIST OF OPTIONS:
      A title description for each dictionary (TITLE) can be
    optionally provided as well as a LABEL for each data point in
    the diagram.

    label = label : label for each data point in target diagram, e.g.
                    'OC445 (CB)'
    overwrite = boolean : true/false flag to overwrite Excel file
    title = title : title descriptor data set, e.g. 'Expt. 01.0'

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Dec 12, 2016
    """
    option = get_write_taylor_stats_options(**kwargs)

    # Check for existence of file
    if os.path.isfile(filename):
        if option["overwrite"]:
            os.remove(filename)
        else:
            ValueError(f"File already exists: {filename}")

    # Convert data to list if necessary
    if not isinstance(data, list):
        data = [data]

    # Write title information to file
    workbook = xlsxwriter.Workbook(filename)
    worksheet = workbook.add_worksheet()

    # Write title information to file
    worksheet.write(1, 0, "Taylor Statistics")

    # Determine number of dictionaries in data variable
    ncell = len(data)

    # Write data for each dictionary
    row = 2
    headers = ["Description", "Standard Deviation", "CRMSD", "Correlation Coeff."]
    for i in range(ncell):
        row += 1
        if len(option["title"]) > 0:
            worksheet.write(row, 0, option["title"][i])

        # Write column headers
        row += 1
        for j, h in enumerate(headers):
            worksheet.write(row, j, h)

        # Retrieve input values as list
        try:
            iter(data[i]["sdev"])
        except TypeError:
            sdev = [data[i]["sdev"]]
            crmsd = [data[i]["crmsd"]]
            ccoef = [data[i]["ccoef"]]
        else:
            sdev = data[i]["sdev"]
            crmsd = data[i]["crmsd"]
            ccoef = data[i]["ccoef"]
        ndata = len(sdev)

        # Write each row of data
        row += 1
        for j in range(ndata):
            if len(option["label"]) > 0:
                worksheet.write(row, 0, option["label"][j])

            worksheet.write(row, 1, sdev[j])
            worksheet.write(row, 2, crmsd[j])
            worksheet.write(row, 3, ccoef[j])
            row += 1

    workbook.close()


def get_write_taylor_stats_options(**kwargs):
    """
    Get optional arguments for write_taylor_stats function.

    Retrieves the keywords supplied to the WRITE_TARGET_STATS
    function (**KWARGS), and returns the values in a OPTION dictionary.
    Default values are assigned to selected optional arguments. The
    function will terminate with an error if an unrecognized optional
    argument is supplied.

    INPUTS:
    **kwargs : keyword argument list

    OUTPUTS:
    option : data structure containing option values.
    option['title']     : title descriptor for data set.
    option['overwrite'] : boolean to overwrite Excel file.

    LIST OF OPTIONS:
      A title description for each dataset TITLE can be optionally
    provided as well as an overwrite option if the file name currently
    exists.

    label = label : label for each data point in target diagram, e.g.
                    'OC445 (CB)'
    overwrite = boolean : true/false flag to overwrite Excel file
    title = title : title descriptor for each data set in data, e.g.
                   'Expt. 01.0'

    Author: Peter A. Rochford
    Acorn Science & Innovation
    prochford@acornsi.com

    Created on Dec 10, 2016

    @author: rochfordp
    """
    from . import check_on_off

    nargin = len(kwargs)

    #  Set default parameters
    option = {"overwrite": False, "label": [], "title": ""}
    if nargin == 0:
        # No options requested, so return with only defaults
        return option

    # Load custom options, storing values in option data structure

    # Check for valid keys and values in dictionary
    for optname, optvalue in kwargs.items():
        optname = optname.lower()
        if optname not in option:
            raise ValueError(f"Unrecognized option: {optname}")
        # Replace option value with that from arguments
        option[optname] = optvalue

        # Check values for specific options
        if optname == "overwrite":
            option["overwrite"] = check_on_off(option["overwrite"])

    return option



"""Clean dist files after tox -e check"""
from pathlib import Path

folder = Path(__file__).resolve().parents[1]

whls = folder.glob("dist/*.whl")

for f in whls:
    f.unlink()

targz = folder.glob("dist/*.tar.gz")

for f in targz:
    f.unlink()



"""Collection of functions for the manipulation of time series."""

import sys

import cltoolbox
import numpy as np
import pandas as pd
from cltoolbox.rst_text_formatter import RSTHelpFormatter
from pydantic import validate_arguments
from toolbox_utils import tsutils


def _dtw(ts_a, ts_b, d=lambda x, y: abs(x - y), window=10000):
    """Return the DTW similarity distance timeseries numpy arrays.

    Parameters
    ----------
    ts_a, ts_b : array of shape [n_samples, n_timepoints]
        Two arrays containing n_samples of timeseries data
        whose DTW distance between each sample of A and B
        will be compared
    d : DistanceMetric object (default = abs(x-y))
        the distance measure used for A_i - B_j in the
        DTW dynamic programming function

    Returns
    -------
    DTW distance between A and B

    """
    # Create cost matrix via broadcasting with large int
    ts_a, ts_b = np.array(ts_a), np.array(ts_b)
    M, N = len(ts_a), len(ts_b)
    cost = sys.maxsize * np.ones((M, N))

    # Initialize the first row and column
    cost[0, 0] = d(ts_a[0], ts_b[0])
    for i in range(1, M):
        cost[i, 0] = cost[i - 1, 0] + d(ts_a[i], ts_b[0])

    for j in range(1, N):
        cost[0, j] = cost[0, j - 1] + d(ts_a[0], ts_b[j])

    # Populate rest of cost matrix within window
    for i in range(1, M):
        for j in range(max(1, i - window), min(N, i + window)):
            choices = cost[i - 1, j - 1], cost[i, j - 1], cost[i - 1, j]
            cost[i, j] = min(choices) + d(ts_a[i], ts_b[j])

    # Return DTW distance given window
    return cost[-1, -1]


@cltoolbox.command("dtw", formatter_class=RSTHelpFormatter)
@tsutils.doc(tsutils.docstrings)
def dtw_cli(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    round_index=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    window=10000,
    source_units=None,
    target_units=None,
    tablefmt="csv",
):
    """Dynamic Time Warping.

    Parameters
    ----------
    window : int
        [optional, default is 10000]

        Window length.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${round_index}
    ${dropna}
    ${skiprows}
    ${index_type}
    ${source_units}
    ${target_units}
    ${names}
    ${clean}
    ${tablefmt}
    """
    tsutils.printiso(
        dtw(
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            round_index=round_index,
            dropna=dropna,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            clean=clean,
            window=window,
            source_units=source_units,
            target_units=target_units,
        ),
        tablefmt=tablefmt,
    )


@validate_arguments
@tsutils.copy_doc(dtw_cli)
def dtw(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    round_index=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    window: int = 10000,
    source_units=None,
    target_units=None,
):
    """Dynamic Time Warping."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )

    process = {}
    for i in tsd.columns:
        for j in tsd.columns:
            if (i, j) not in process and (j, i) not in process and i != j:
                process[(i, j)] = _dtw(tsd[i], tsd[j], window=window)

    ntsd = pd.DataFrame(list(process.items()))
    ncols = ntsd.columns
    ncols = ["Variables"] + [f"{str(i)}DTW_score" for i in ncols[1:]]
    ntsd.columns = ncols
    return ntsd



import shlex
import subprocess
from unittest import TestCase

import pandas
from pandas.testing import assert_frame_equal
from toolbox_utils import tsutils

from tstoolbox import tstoolbox


class TestRead(TestCase):
    def setUp(self):
        dr = pandas.date_range("2000-01-01", periods=2, freq="D")

        ts = pandas.Series([4.5, 4.6], index=dr)

        self.read_direct = pandas.DataFrame(ts, columns=["Value"])
        self.read_direct.index.name = "Datetime"
        self.read_direct = tsutils.memory_optimize(self.read_direct)

        self.read_multiple_direct = pandas.DataFrame(ts, columns=["Value"])
        self.read_multiple_direct = pandas.concat(
            [self.read_multiple_direct, pandas.Series(ts, name="Value_r")],
            axis="columns",
        )
        self.read_multiple_direct.index.name = "Datetime"
        self.read_multiple_direct = tsutils.memory_optimize(self.read_multiple_direct)

        self.read_cli = b"""Datetime,Value
2000-01-01,4.5
2000-01-02,4.6
"""

        self.read_multiple_cli = b"""Datetime,Value,Value_r
2000-01-01,4.5,4.5
2000-01-02,4.6,4.6
"""

        self.read_tsstep_2_daily_cli = b"""Datetime,Value,Value1
2000-01-01,4.5,45.6
2000-01-03,4.7,34.2
2000-01-05,4.5,7.2
"""
        self.read_tsstep_2_daily = pandas.DataFrame(
            [[4.5, 45.6], [4.7, 34.2], [4.5, 7.2]],
            columns=["Value", "Value1"],
            index=pandas.DatetimeIndex(["2000-01-01", "2000-01-03", "2000-01-05"]),
        )
        self.read_tsstep_2_daily = tsutils.memory_optimize(self.read_tsstep_2_daily)

        self.read_tsstep_2_daily.index.name = "Datetime"

        self.read_blanks = tsutils.read_iso_ts(
            b"""Datetime,Value::mean,Unnamed::mean,Unnamed.001::mean,Unnamed.002::mean,Unnamed.003::mean,Unnamed.004::mean,Unnamed.005::mean,Unnamed.006::mean,Unnamed.007::mean
2000-01-01,2.46667,,,,,,,,
2000-01-02,3.4,,,,,,,,
"""
        )

    def test_read_direct(self):
        """Test read API for single column - daily."""
        out = tstoolbox.read("tests/data_simple.csv")
        assert_frame_equal(out, self.read_direct)

    def test_read_mulitple_direct(self):
        """Test read API for multiple columns - daily."""
        out = tstoolbox.read(
            "tests/data_simple.csv tests/data_simple.csv", append=r"columns"
        )
        assert_frame_equal(out, self.read_multiple_direct)

    def test_read_mulitple_direct_list(self):
        """Test read API for multiple columns - daily."""
        out = tstoolbox.read(
            ["tests/data_simple.csv", "tests/data_simple.csv"], append=r"columns"
        )
        assert_frame_equal(out, self.read_multiple_direct)

    def test_read_bi_monthly(self):
        """Test read API for bi monthly time series."""
        out = tstoolbox.read("tests/data_bi_daily.csv")
        assert_frame_equal(out, self.read_tsstep_2_daily)

    def test_read_cli(self):
        """Test read CLI for single column - daily."""
        args = "tstoolbox read tests/data_simple.csv"
        args = shlex.split(args)
        out = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()
        self.assertEqual(out[0], self.read_cli)

    def test_read_multiple_cli(self):
        """Test read CLI for multiple columns - daily."""
        args = "tstoolbox read --append columns tests/data_simple.csv tests/data_simple.csv"
        args = shlex.split(args)
        out = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()
        self.assertEqual(out[0], self.read_multiple_cli)

    def test_read_multiple_cli_space(self):
        """Test read CLI for multiple columns - daily."""
        args = "tstoolbox read --append columns tests/data_simple.csv tests/data_simple.csv"
        args = shlex.split(args)
        out = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()
        self.assertEqual(out[0], self.read_multiple_cli)

    def test_read_bi_monthly_cli(self):
        """Test read CLI for bi monthly time series."""
        args = "tstoolbox read tests/data_bi_daily.csv"
        args = shlex.split(args)
        out = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()
        self.assertEqual(out[0], self.read_tsstep_2_daily_cli)

    def test_read_blank_header_cli(self):
        """Test reading of files with blank titles in header."""
        args = (
            "tstoolbox aggregate --agg_interval D --input_ts tests/data_empty_cols.csv"
        )
        args = shlex.split(args)
        out = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()[0]
        out = tsutils.memory_optimize(tsutils.read_iso_ts(out))
        assert_frame_equal(out, tsutils.memory_optimize(self.read_blanks))

    def test_read_multiple_spaces(self):
        """Test reading of files with multiple spaces in data."""
        args = "tstoolbox aggregate --agg_interval D --input_ts tests/data_spaces.csv"
        args = shlex.split(args)
        out = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()[0]
        out = tsutils.read_iso_ts(out)
        assert_frame_equal(out, self.read_blanks)


def test_read_xlsx():
    """Test read API for xlsx"""
    out = tstoolbox.read("tests/data_flow_stage.xlsx")
    comp = tstoolbox.read("tests/data.wdm,2")
    comp.columns = ["Lake Helen"]
    assert_frame_equal(out, comp, check_dtype=False)


def test_read_xlsx_sheet_name():
    """Test read API for xlsx"""
    out = tstoolbox.read("tests/data_flow_stage.xlsx,in")
    comp = tstoolbox.read("tests/data.wdm,2")
    comp.columns = ["in_Lake Helen"]
    assert_frame_equal(out, comp, check_dtype=False)


def test_read_xlsx_sheet_number():
    """Test read API for xlsx"""
    out = pandas.DataFrame(tstoolbox.read("tests/data_flow_stage.xlsx,0,in").iloc[:, 0])
    comp = tstoolbox.read("tests/data.wdm,2")
    comp.columns = ["0_Lake Helen"]
    assert_frame_equal(out, comp, check_dtype=False)



"""Collection of functions for the manipulation of time series."""

import warnings

import cltoolbox
import numpy as np
from cltoolbox.rst_text_formatter import RSTHelpFormatter
from pydantic import validate_arguments
from toolbox_utils import tsutils

warnings.filterwarnings("ignore")


@cltoolbox.command("add_trend", formatter_class=RSTHelpFormatter)
@tsutils.doc(tsutils.docstrings)
def add_trend_cli(
    start_offset,
    end_offset,
    start_index=0,
    end_index=-1,
    input_ts="-",
    start_date=None,
    end_date=None,
    skiprows=None,
    columns=None,
    clean=False,
    dropna="no",
    names=None,
    source_units=None,
    target_units=None,
    round_index=None,
    index_type="datetime",
    print_input=False,
    tablefmt="csv",
):
    """Add a trend.

    Adds a linear interpolated trend to the input data.  The trend
    values start at [`start_index`, `start_offset`] and end at
    [`end_index`, `end_offset`].

    Parameters
    ----------
    start_offset : float
        The starting value for the applied trend.  This is the starting
        value for the linear interpolation that will be added to the
        input data.
    end_offset : float
        The ending value for the applied trend.  This is the ending
        value for the linear interpolation that will be added to the
        input data.
    start_index : int
        [optional, default is 0, transformation]

        The starting index where `start_offset` will be initiated.  Rows
        prior to `start_index` will not be affected.
    end_index : int
        [optional, default is -1, transformation]

        The ending index where `end_offset` will be set.  Rows after
        `end_index` will not be affected.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${clean}
    ${round_index}
    ${skiprows}
    ${index_type}
    ${names}
    ${source_units}
    ${target_units}
    ${print_input}
    ${tablefmt}
    """
    tsutils.printiso(
        add_trend(
            start_offset,
            end_offset,
            start_index=start_index,
            end_index=end_index,
            input_ts=input_ts,
            columns=columns,
            clean=clean,
            start_date=start_date,
            end_date=end_date,
            dropna=dropna,
            round_index=round_index,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            source_units=source_units,
            target_units=target_units,
            print_input=print_input,
        ),
        tablefmt=tablefmt,
    )


@validate_arguments
@tsutils.copy_doc(add_trend_cli)
def add_trend(
    start_offset: float,
    end_offset: float,
    start_index: int = 0,
    end_index: int = -1,
    input_ts="-",
    columns=None,
    clean=False,
    start_date=None,
    end_date=None,
    dropna="no",
    round_index=None,
    skiprows=None,
    index_type="datetime",
    names=None,
    source_units=None,
    target_units=None,
    print_input=False,
):
    """Add a trend."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )
    # Need it to be float since will be using np.nan
    ntsd = tsd.copy().astype("float64")

    ntsd.iloc[:, :] = np.nan
    ntsd.iloc[start_index, :] = start_offset
    ntsd.iloc[end_index, :] = end_offset
    ntsd = ntsd.interpolate(method="values")

    ntsd = ntsd + tsd

    ntsd = tsutils.memory_optimize(ntsd)
    return tsutils.return_input(print_input, tsd, ntsd, "trend")



__version__ = "1.1.4"

from .apc_bias import apc_bias
from .bias import bias
from .brier_score import brier_score
from .centered_rms_dev import centered_rms_dev
from .check_duplicate_stats import check_duplicate_stats
from .check_on_off import check_on_off
from .check_taylor_stats import check_taylor_stats
from .get_target_diagram_axes import get_target_diagram_axes
from .get_target_diagram_options import get_target_diagram_options
from .get_taylor_diagram_axes import get_taylor_diagram_axes
from .get_taylor_diagram_options import get_taylor_diagram_options
from .index_agreement import index_agreement
from .kge09 import kge09

kge = kge09
from .kge12 import kge12
from .nse import nse
from .overlay_target_diagram_circles import overlay_target_diagram_circles
from .overlay_taylor_diagram_circles import overlay_taylor_diagram_circles
from .overlay_taylor_diagram_lines import overlay_taylor_diagram_lines
from .pc_bias import pc_bias
from .plot_pattern_diagram_colorbar import plot_pattern_diagram_colorbar
from .plot_pattern_diagram_markers import plot_pattern_diagram_markers
from .plot_target_axes import plot_target_axes
from .plot_taylor_axes import plot_taylor_axes
from .plot_taylor_obs import plot_taylor_obs
from .report_duplicate_stats import report_duplicate_stats
from .rmsd import rmsd
from .skill_score_brier import skill_score_brier
from .skill_score_murphy import skill_score_murphy
from .target_diagram import target_diagram
from .target_statistics import target_statistics
from .taylor_diagram import taylor_diagram
from .taylor_statistics import taylor_statistics
from .write_stats import write_stats
from .write_target_stats import write_target_stats
from .write_taylor_stats import write_taylor_stats



from array import array

import numpy as np

from . import (
    check_taylor_stats,
    get_taylor_diagram_axes,
    get_taylor_diagram_options,
    overlay_taylor_diagram_circles,
    overlay_taylor_diagram_lines,
    plot_pattern_diagram_colorbar,
    plot_pattern_diagram_markers,
    plot_taylor_axes,
    plot_taylor_obs,
)


def taylor_diagram(*args, **kwargs):
    """
    Plot a Taylor diagram from statistics of different series.

    taylor_diagram(stds,rmss,cors,keyword=value)

    The first 3 arguments must be the inputs as described below followed by
    keywords in the format OPTION = value. An example call to the function
    would be:

    taylor_diagram(stds,rmss,cors,markerdisplayed='marker')

    INPUTS:
    stds: Standard deviations
    rmss: Centered Root Mean Square Difference
    cors: Correlation

    Each of these inputs are one-dimensional with the same length. First
    index corresponds to the reference series for the diagram. For
    example stds[1] is the standard deviation of the reference series
    and stds[2:N] are the standard deviations of the other series. Note
    that only the latter are plotted.

    Note that by definition the following relation must be true for all
    series i:

     rmss(i) = sqrt(stds(i).^2 + stds(1)^2 - 2*stds(i)*stds(1).*cors(i))

    This relation is checked if the checkStats option is used, and if not
    verified an error message is sent. This relation is not checked by
    default. Please see Taylor's JGR article for more informations about
    this relation.

    OUTPUTS:
    None.

    LIST OF OPTIONS:
    For an exhaustive list of options to customize your diagram, call the
    function without arguments at a Python command line:
    % python
    >> from tstoolbox.skill_metrics.taylor_diagram import taylor_diagram
    >> taylor_diagram()

    Reference:

    Taylor, K. E. (2001), Summarizing multiple aspects of model
      performance in a single diagram, J. Geophys. Res., 106(D7),
      7183-7192, doi:10.1029/2000JD900719.

    Author: Peter A. Rochford
            Symplectic, LLC
            www.thesymplectic.com
            prochford@thesymplectic.com

    Created on Dec 3, 2016
    """
    # Check for number of arguments
    nargin = len(args)
    stds, rmss, cors = _get_taylor_diagram_arguments(*args, **kwargs)
    if nargin == 0:
        return

    # Get options
    option = get_taylor_diagram_options(cors, **kwargs)

    # Check the input statistics if requested.
    if option["checkstats"] == "on":
        check_taylor_stats(stds[1:], rmss[1:], cors[1:], 0.01)

    # Express statistics in polar coordinates.
    rho = stds
    theta = np.arccos(cors)

    #  Get axis values for plot
    axes, cax = get_taylor_diagram_axes(rho, option)

    # Plot axes for target diagram
    if option["overlay"] == "off":
        # Draw circles about origin
        overlay_taylor_diagram_circles(axes, cax, option)

        # Draw lines emanating from origin
        overlay_taylor_diagram_lines(axes, cax, option)

        # Plot axes for Taylor diagram
        ax = plot_taylor_axes(axes, cax, option)

        # Plot marker on axis indicating observation STD
        plot_taylor_obs(ax, stds[0], axes, option)

    # Plot data points. Note that only rho[1:N] and theta[1:N] are
    # plotted.
    x = np.multiply(rho[1:], np.cos(theta[1:]))
    y = np.multiply(rho[1:], np.sin(theta[1:]))

    lowcase = option["markerdisplayed"].lower()
    if lowcase == "marker":
        plot_pattern_diagram_markers(x, y, option)
    elif lowcase == "colorbar":
        plot_pattern_diagram_colorbar(x, y, rmss[1:], option)
    else:
        raise ValueError("Unrecognized option: option['markerdisplayed']")


def _get_taylor_diagram_arguments(*args, **kwargs):
    """
    Get arguments for taylor_diagram function.

    Retrieves the arguments supplied to the TAYLOR_DIAGRAM function as
    arguments and displays the optional arguments if none are supplied.
    Otherwise, tests the first 3 arguments are numeric quantities and
    returns their values.

    INPUTS:
    args : variable-length input argument list

    OUTPUTS:
    stds: Standard deviations
    rmss: Centered Root Mean Square Difference
    cors: Correlation
    """
    import numbers

    stds = []
    rmss = []
    cors = []

    nargin = len(args)
    if nargin == 0:
        # Display options list
        _display_taylor_diagram_options()
        return stds, rmss, cors
    if nargin != 3:
        raise ValueError("Must supply 3 arguments.")

    stds = args[0]
    rmss = args[1]
    cors = args[2]

    # Test the above are numeric quantities
    if isinstance(stds, array):
        stds = np.array(stds)
    if isinstance(stds, numbers.Number):
        stds = np.array(stds, ndmin=1)
    if not isinstance(stds, np.ndarray):
        raise ValueError("Argument stds is not a numeric array")

    if isinstance(rmss, array):
        rmss = np.array(rmss)
    if isinstance(rmss, numbers.Number):
        rmss = np.array(rmss, ndmin=1)
    if not isinstance(rmss, np.ndarray):
        raise ValueError("Argument rmss is not a numeric array")

    if isinstance(cors, array):
        cors = np.array(cors)
    if isinstance(cors, numbers.Number):
        cors = np.array(cors, ndmin=1)
    if not isinstance(cors, (np.ndarray, array)):
        raise ValueError("Argument cors is not a numeric array")

    return stds, rmss, cors


def _display_taylor_diagram_options():
    """Display available options for TAYLOR_DIAGRAM function."""
    _disp("General options:")
    _dispopt(
        "'numberPanels'",
        """1 or 2: Panels to display
(1 for positive correlations, 2 for positive and negative correlations).
    Default value depends on correlations (cors)""",
    )
    _dispopt(
        "'overlay'",
        """'on' / 'off' (default):
        Switch to overlay current statistics on Taylor diagram.
        Only markers will be displayed.""",
    )
    _dispopt(
        "'alpha'",
        """Blending of symbol face color (0.0 transparent through 1.0 opaque)
(Default: 1.0)""",
    )
    _dispopt("'axismax'", "Maximum for the radial contours")
    _dispopt(
        "'colormap'",
        """'on'/ 'off' (default):
        Switch to map color shading of markers to colormap ('on')
        or min to max range of RMSDz values ('off').
        Set to same value as option['nonRMSDz'].""",
    )
    _disp("")

    _disp("Marker options:")
    _dispopt(
        "'MarkerDisplayed'",
        """'marker' (default):
        Experiments are represented by individual symbols
        "'colorBar':
             Experiments are represented by a color described in a colorbar""",
    )
    _disp("OPTIONS when 'MarkerDisplayed' == 'marker'")
    _dispopt("'markerLabel'", "Labels for markers")
    _dispopt("'markerLabelColor'", "Marker label color (Default: black)")
    _dispopt("'markerColor'", "Single color to use for all markers (Default: red)")
    _dispopt("'markerLegend'", "'on' / 'off' (default): Use legend for markers")
    _dispopt("'markerSize'", "Marker size (Default: 10)")

    _disp("OPTIONS when MarkerDisplayed' == 'colorbar'")
    _dispopt(
        "'nonRMSDz'",
        """'on'/ 'off' (default):
        Values in RMSDz do not correspond to total RMS Differences.
        (Used to make range of RMSDz values appear above color bar.)""",
    )
    _dispopt("'titleColorBar'", "Title of the colorbar.")
    _disp("")

    _disp("RMS axis options:")
    _dispopt("'tickRMS'", "RMS values to plot grid circles from observation point")
    _dispopt("'rincRMS'", "axis tick increment for RMS values")
    _dispopt("'colRMS'", "RMS grid and tick labels color. (Default: green)")
    _dispopt("'showlabelsRMS'", "'on' (default) / 'off': Show the RMS tick labels")
    _dispopt(
        "'tickRMSangle'",
        "Angle for RMS tick labels with the observation point. Default: 135 deg.",
    )
    _dispopt(
        "'rmsLabelFormat'",
        """String format for RMS contour labels, e.g. '0:.2f'.
        (Default '0', format as specified by str function.)""",
    )
    _dispopt("'styleRMS'", "Line style of the RMS grid")
    _dispopt("'widthRMS'", "Line width of the RMS grid")
    _dispopt("'titleRMS'", "'on' (default) / 'off': Show RMSD axis title")
    _disp("")

    _disp("STD axis options:")
    _dispopt("'tickSTD'", "STD values to plot gridding circles from origin")
    _dispopt("'rincSTD'", "axis tick increment for STD values")
    _dispopt("'colSTD'", "STD grid and tick labels color. (Default: black)")
    _dispopt("'showlabelsSTD'", "'on' (default) / 'off': Show the STD tick labels")
    _dispopt("'styleSTD'", "Line style of the STD grid")
    _dispopt("'widthSTD'", "Line width of the STD grid")
    _dispopt("'titleSTD'", "'on' (default) / 'off': Show STD axis title")
    _disp("")

    _disp("CORRELATION axis options:")
    _dispopt("'tickCOR'", "CORRELATON grid values")
    _dispopt("'colCOR'", "CORRELATION grid color. Default: blue")
    _dispopt(
        "'showlabelsCOR'", "'on' (default) / 'off': Show the CORRELATION tick labels"
    )
    _dispopt("'styleCOR'", "Line style of the COR grid")
    _dispopt("'widthCOR'", "Line width of the COR grid")
    _dispopt("'titleCOR'", "'on' (default) / 'off': Show CORRELATION axis title")
    _disp("")

    _disp("Observation Point options:")
    _dispopt("'colObs'", "Observation STD color. (Default: magenta)")
    _dispopt(
        "'markerObs'",
        """Marker to use for x-axis indicating observed STD.
        A choice of 'none' will suppress appearance of marker. (Default 'none')""",
    )
    _dispopt(
        "'styleObs'",
        """Line style for observation grid line. A choice of empty string ('')
        will suppress appearance of the grid line. (Default: '')""",
    )
    _dispopt("'titleOBS'", "Label for observation point (Default: '')")
    _dispopt("'widthOBS'", "Line width for observation grid line")
    _disp("")

    _disp("CONTROL options:")
    _dispopt(
        "'checkStats'",
        """'on' / 'off' (default):
        Check input statistics satisfy Taylor relationship""",
    )


def _disp(text):
    print(text)


def _dispopt(optname, optval):
    """
    Display option name and values.

    This is a support function for the DISPLAY_TARGET_DIAGRAM_OPTIONS function.
    It displays the option name OPTNAME on a line by itself followed by its
    value OPTVAL on the following line.
    """
    _disp(f"\t{optname}")
    _disp(f"\t\t{optval}")



import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt
import pandas as pd
import pytest
from toolbox_utils import tsutils

from tstoolbox import tstoolbox

# Pull this in once.
idf = tsutils.common_kwds(input_tsd="tests/02234500_65_65.csv", clean=True)

df = idf.resample("D").agg("mean")

dfa = idf.resample("A").agg("mean")


@pytest.mark.mpl_image_compare(tolerance=6)
def test_histogram():
    plt.close("all")
    return tstoolbox.plot(
        type="histogram",
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        sharex=False,
        plot_styles="classic",
    )


# @pytest.mark.mpl_image_compare(tolerance=6)
# def test_heatmap():
#     plt.close("all")
#     return tstoolbox.heatmap(
#         columns=2,
#         clean=True,
#         input_ts=df,
#         ofilename=None,
#         plot_styles="classic",
#     )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_norm_xaxis():
    plt.close("all")
    return tstoolbox.plot(
        type="norm_xaxis",
        columns=2,
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_norm_yaxis():
    plt.close("all")
    return tstoolbox.plot(
        type="norm_yaxis",
        columns=2,
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_lognorm_xaxis():
    plt.close("all")
    return tstoolbox.plot(
        type="lognorm_xaxis",
        columns=2,
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_lognorm_yaxis():
    plt.close("all")
    return tstoolbox.plot(
        type="lognorm_yaxis",
        columns=2,
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_weibull_xaxis():
    plt.close("all")
    return tstoolbox.plot(
        type="weibull_xaxis",
        columns=2,
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_weibull_yaxis():
    plt.close("all")
    return tstoolbox.plot(
        type="weibull_yaxis",
        columns=2,
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_kde_time():
    plt.close("all")
    return tstoolbox.plot(
        type="kde_time",
        columns=2,
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_kde_time_multiple_traces():
    plt.close("all")
    ndf = pd.read_csv("tests/data_daily.csv", index_col=0, parse_dates=True)
    ndf = ndf.join(
        pd.read_csv("tests/data_02325000_flow.csv", index_col=0, parse_dates=True),
        how="outer",
    )
    return tstoolbox.plot(
        type="kde_time",
        columns=[2, 3],
        clean=True,
        input_ts=ndf,
        ytitle="Flow",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_autocorrelation():
    plt.close("all")
    return tstoolbox.plot(
        type="autocorrelation",
        columns=2,
        input_ts=df,
        ofilename=None,
        plot_styles="classic",
    )


# @pytest.mark.mpl_image_compare(tolerance=6)
# def test_taylor():
#     plt.close("all")
#     return tstoolbox.taylor(
#         clean=True,
#         input_ts="tests/02234500_65_65.csv",
#         ofilename=None,
#         plot_styles="classic",
#     )
#
#
# @pytest.mark.mpl_image_compare(tolerance=6)
# def test_target():
#     plt.close("all")
#     return tstoolbox.target(
#         clean=True,
#         input_ts="tests/02234500_65_65.csv",
#         ofilename=None,
#         plot_styles="classic",
#     )



"""Collection of functions for the manipulation of time series."""

import warnings
from typing import List, Optional

import cltoolbox
import pandas as pd
from cltoolbox.rst_text_formatter import RSTHelpFormatter
from pydantic import confloat, conint, validate_arguments
from toolbox_utils import tsutils

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


warnings.filterwarnings("ignore")


@cltoolbox.command("ewm_window", formatter_class=RSTHelpFormatter)
@tsutils.doc(tsutils.docstrings)
def ewm_window_cli(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    statistic="",
    alpha_com=None,
    alpha_span=None,
    alpha_halflife=None,
    alpha=None,
    min_periods=0,
    adjust=True,
    ignore_na=False,
    source_units=None,
    target_units=None,
    print_input=False,
    tablefmt="csv",
):
    """Calculate exponential weighted functions.

    Exactly one of `alpha_com` (center of mass), `alpha_span`,
    `alpha_halflife`, and `alpha` must be provided to calculate the
    'alpha' term.  Allowed values and relationship between the
    parameters are specified in the parameter descriptions below; see
    the link at the end of this section for a detailed explanation.

    When `adjust` is True (default), weighted averages are calculated
    using weights ``(1-alpha)**(n-1), (1-alpha)**(n-2), . . . , 1-alpha, 1``.

    When `adjust` is False, weighted averages are calculated recursively
    as:
    ``weighted_average[0] = arg[0]``
    ``weighted_average[i] = (1-alpha)*weighted_average[i-1] + alpha*arg[i]``

    When `ignore_na` is False (default), weights are based on absolute
    positions.  For example, the weights of x and y used in calculating
    the final weighted average of [x, None, y] are ``(1-alpha)**2`` and
    1 (if `adjust` is True), and ``(1-alpha)**2`` and alpha (if `adjust` is
    False).

    When `ignore_na` is True weights are based on relative positions.
    For example, the weights of x and y used in calculating the final
    weighted average of [x, None, y] are 1-alpha and 1 (if `adjust` is
    True), and 1-alpha and alpha (if adjust is `False`).

    More details can be found at
    http://pandas.pydata.org/pandas-docs/stable/computation.html#exponentially-weighted-windows

    Parameters
    ----------
    statistic : str
        [optional, defaults to '']

        Statistic applied to each window.

        +-----------+--------------------+
        | statistic | Description        |
        +===========+====================+
        | corr      | correlation        |
        +-----------+--------------------+
        | cov       | covariance         |
        +-----------+--------------------+
        | mean      | mean               |
        +-----------+--------------------+
        | std       | standard deviation |
        +-----------+--------------------+
        | var       | variance           |
        +-----------+--------------------+
    alpha_com : float
        [optional, defaults to None]

        Specify decay in terms of center of mass::

            alpha = 1/(1+`alpha_com`), for `alpha_com` >= 0
    alpha_span : float
        [optional, defaults to None]

        Specify decay in terms of span::

            alpha = 2/(`alpha_span`+1), for `alpha_span` > 1
    alpha_halflife : float
        [optional, defaults to None]

        Specify decay in terms of half-life::

            alpha = 1-exp(log(0.5)/alpha_halflife), for
            alpha_halflife > 0
    alpha : float
        [optional, defaults to None]

        Specify smoothing factor ``alpha`` directly, ``0<alpha<=1``
    min_periods : int
        [optional, default is 0]

        Minimum number of observations in window required to have a value
        (otherwise result is NA).
    adjust : boolean
        [optional, default is True]

        Divide by decaying adjustment factor in beginning periods to account
        for imbalance in relative weightings (viewing EWMA as a moving average)
    ignore_na : boolean
        [optional, default is False]
        Ignore missing values when calculating weights.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${skiprows}
    ${index_type}
    ${names}
    ${clean}
    ${source_units}
    ${target_units}
    ${print_input}
    ${tablefmt}
    """
    tsutils.printiso(
        ewm_window(
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            dropna=dropna,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            clean=clean,
            statistic=statistic,
            alpha_com=alpha_com,
            alpha_span=alpha_span,
            alpha_halflife=alpha_halflife,
            alpha=alpha,
            min_periods=min_periods,
            adjust=adjust,
            ignore_na=ignore_na,
            source_units=source_units,
            target_units=target_units,
            print_input=print_input,
        ),
        tablefmt=tablefmt,
    )


@tsutils.transform_args(statistic=tsutils.make_list)
@validate_arguments
@tsutils.copy_doc(ewm_window_cli)
def ewm_window(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    statistic: List[Optional[Literal["corr", "cov", "mean", "std", "var"]]] = "mean",
    alpha_com: Optional[confloat(ge=0)] = None,
    alpha_span: Optional[confloat(ge=0)] = None,
    alpha_halflife: Optional[confloat(ge=0)] = None,
    alpha: Optional[confloat(ge=0, le=1)] = None,
    min_periods: Optional[conint(ge=0)] = 0,
    adjust: bool = True,
    ignore_na: bool = False,
    source_units=None,
    target_units=None,
    print_input=False,
):
    """Calculate exponential weighted functions."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )

    ntsd = tsd.ewm(
        alpha_com=alpha_com,
        alpha_span=alpha_span,
        alpha_halflife=alpha_halflife,
        alpha=alpha,
        min_periods=min_periods,
        adjust=adjust,
        ignore_na=ignore_na,
    )

    if statistic:
        nntsd = pd.DataFrame()
        for stat in statistic:
            ntsd = eval(f"ntsd.{stat}()")
            ntsd = [tsutils.renamer(i, f"ewm.{stat}") for i in ntsd.columns]
            nntsd = nntsd.join(ntsd, how="outer")
    else:
        nntsd = ntsd

    return tsutils.return_input(print_input, tsd, nntsd)



"""Collection of functions for the manipulation of time series."""

import warnings

import cltoolbox
import pandas as pd
from cltoolbox.rst_text_formatter import RSTHelpFormatter
from pydantic import validate_arguments
from toolbox_utils import tsutils

warnings.filterwarnings("ignore")


@cltoolbox.command("calculate_kde", formatter_class=RSTHelpFormatter)
@tsutils.doc(tsutils.docstrings)
def calculate_kde_cli(
    ascending=True,
    evaluate=False,
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    clean=False,
    skiprows=None,
    index_type="datetime",
    source_units=None,
    target_units=None,
    names=None,
    tablefmt="csv",
):
    """Return the kernel density estimation (KDE) curve.

    Returns a time-series or the KDE curve depending on the `evaluate`
    keyword.

    Parameters
    ----------
    ascending : bool
        [optional, defaults to True, input filter]

        Sort order.
    evaluate : bool
        [optional, defaults to False, transformation]

        Whether or not to return a time-series of KDE density values or
        the KDE curve.  Defaults to False, which would return the KDE
        curve.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${skiprows}
    ${index_type}
    ${names}
    ${source_units}
    ${target_units}
    ${clean}
    ${tablefmt}
    """
    tsutils.printiso(
        calculate_kde(
            ascending=ascending,
            evaluate=evaluate,
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            clean=clean,
            skiprows=skiprows,
            index_type=index_type,
            source_units=source_units,
            target_units=target_units,
            names=names,
        ),
        tablefmt=tablefmt,
    )


@validate_arguments
@tsutils.copy_doc(calculate_kde_cli)
def calculate_kde(
    input_ts="-",
    ascending: bool = True,
    evaluate: bool = False,
    columns=None,
    start_date=None,
    end_date=None,
    clean=False,
    skiprows=None,
    index_type="datetime",
    source_units=None,
    target_units=None,
    names=None,
):
    """Return the kernel density estimation (KDE) curve."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )

    if len(tsd.columns) > 1:
        raise ValueError(
            tsutils.error_wrapper(
                f"""
Right now "calculate_kde" only support one time-series at a time.

You gave {tsd.columns}.
"""
            )
        )

    from scipy.stats import gaussian_kde

    tmptsd = tsd.dropna()
    ndf = tmptsd.sort_values(tmptsd.columns[0], ascending=ascending)
    gkde = gaussian_kde(ndf.iloc[:, 0])

    if evaluate:
        y = gkde.evaluate(tmptsd.iloc[:, 0])
        ndf = pd.DataFrame(y, index=tmptsd.index)
    else:
        y = gkde.evaluate(ndf.iloc[:, 0])
        ndf = pd.DataFrame(y)

    return ndf



"""Collection of functions for the manipulation of time series."""


import warnings
from typing import List, Optional

import cltoolbox
import pandas as pd
from cltoolbox.rst_text_formatter import RSTHelpFormatter
from pydantic import conint, validate_arguments
from toolbox_utils import tsutils

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


warnings.filterwarnings("ignore")


@cltoolbox.command("rolling_window", formatter_class=RSTHelpFormatter)
@tsutils.doc(tsutils.docstrings)
def rolling_window_cli(
    statistic,
    groupby=None,
    window=None,
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    span=None,
    min_periods=None,
    center=False,
    win_type=None,
    on=None,
    closed=None,
    source_units=None,
    target_units=None,
    print_input=False,
    tablefmt="csv",
):
    """Calculate a rolling window statistic.

    Parameters
    ----------
    statistic : str
        The statistic that will be applied to each
        window.

        +----------+--------------------+
        | corr     | correlation        |
        +----------+--------------------+
        | count    | count of numbers   |
        +----------+--------------------+
        | cov      | covariance         |
        +----------+--------------------+
        | kurt     | kurtosis           |
        +----------+--------------------+
        | max      | maximum            |
        +----------+--------------------+
        | mean     | mean               |
        +----------+--------------------+
        | median   | median             |
        +----------+--------------------+
        | min      | minimum            |
        +----------+--------------------+
        | quantile | quantile           |
        +----------+--------------------+
        | skew     | skew               |
        +----------+--------------------+
        | std      | standard deviation |
        +----------+--------------------+
        | sum      | sum                |
        +----------+--------------------+
        | var      | variance           |
        +----------+--------------------+
    ${groupby}
    window
        [optional, default = 2]

        Size of the moving window. This is the number of observations used for
        calculating the statistic. Each window will be a fixed size.

        If it is an offset then this will be the time period of each window. Each
        window will be a variable sized based on the observations included in
        the time-period. This is only valid for datetimelike indexes.
    min_periods : int
        [optional, default is None]

        Minimum number of observations in window required to have a value
        (otherwise result is NA). For a window that is specified by an offset,
        this will default to 1.
    center : boolean
        [optional, default is False]

        Set the labels at the center of the window.
    win_type : str
        [optional, default is None]

        Provide a window type.

        One of::

            boxcar
            triang
            blackman
            hamming
            bartlett
            parzen
            bohman
            blackmanharris
            nuttall
            barthann
            kaiser (needs beta)
            gaussian (needs std)
            general_gaussian (needs power, width)
            slepian (needs width)
            exponential (needs tau), center is set to None.
    on : str
        [optional, default is None]

        For a DataFrame, column on which to calculate the rolling window,
        rather than the index
    closed : str
        [optional, default is None]

        Make the interval closed on the 'right', 'left', 'both' or 'neither'
        endpoints. For offset-based windows, it defaults to 'right'. For fixed
        windows, defaults to 'both'. Remaining cases not implemented for fixed
        windows.
    span :
        [optional, default = 2]

        DEPRECATED: Changed to 'window' to be consistent with pandas.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${skiprows}
    ${index_type}
    ${names}
    ${clean}
    ${source_units}
    ${target_units}
    ${print_input}
    ${tablefmt}
    """
    tsutils.printiso(
        rolling_window(
            statistic,
            groupby=groupby,
            window=window,
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            dropna=dropna,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            clean=clean,
            span=span,
            min_periods=min_periods,
            center=center,
            win_type=win_type,
            on=on,
            closed=closed,
            source_units=source_units,
            target_units=target_units,
            print_input=print_input,
        ),
        tablefmt=tablefmt,
    )


@tsutils.transform_args(window=tsutils.make_list)
@validate_arguments
@tsutils.copy_doc(rolling_window_cli)
def rolling_window(
    statistic: Literal[
        "corr",
        "count",
        "cov",
        "kurt",
        "max",
        "mean",
        "median",
        "min",
        "quantile",
        "skew",
        "std",
        "sum",
        "var",
    ],
    groupby=None,
    window: Optional[List[conint(ge=0)]] = None,
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    span=None,
    min_periods: Optional[conint(ge=0)] = None,
    center: bool = False,
    win_type: Optional[str] = None,
    on: Optional[str] = None,
    closed: Optional[Literal["right", "left", "both", "neither"]] = None,
    source_units=None,
    target_units=None,
    print_input=False,
):
    """Calculate a rolling window statistic."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
        groupby=groupby,
    )

    if span is not None:
        window = [span]
    if window is None:
        window = [2]

    ntsd = pd.DataFrame()
    for win in window:
        statstr = ""
        if statistic in [
            "corr",
            "count",
            "cov",
            "kurt",
            "max",
            "mean",
            "median",
            "min",
            "quantile",
            "skew",
            "std",
            "sum",
            "var",
        ]:
            statstr = f".{statistic}()"
        etsd = eval(
            f"""tsd.apply(lambda x:
                       x.rolling({win},
                                 min_periods={min_periods},
                                 center={center},
                                 win_type={win_type},
                                 on={on},
                                 closed={closed}){statstr}
                                 )"""
        )
        etsd.columns = [
            tsutils.renamer(i, f"rolling.{win}.{statistic}") for i in etsd.columns
        ]

        ntsd = ntsd.join(etsd, how="outer")

    return tsutils.return_input(print_input, tsd, ntsd, None)



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Filling
# ## Fill forward

# %%
import pandas as pd

import tstoolbox.tstoolbox as ts

# %%
dr = pd.date_range(periods=16, start="2010-01-01", freq="H")
df = pd.DataFrame([1.1, 2.3, 2.1] + 10 * [None] + [5.4, 6.5, 7.2], dr)

# %%
df

# %%
ts.plot(input_ts=df, ofilename="base.png", style="r-*")

# %% [markdown]
# ![](base.png)

# %%
ndf = ts.fill(input_ts=df)
ndf

# %%
ts.plot(input_ts=ndf, ofilename="ffill.png", style="r-*")

# %% [markdown]
# ![](ffill.png)

# %%
ndf = ts.fill(input_ts=df, method="bfill")

# %%
ts.plot(input_ts=ndf, ofilename="bfill.png", style="r-*")

# %% [markdown]
# ![](bfill.png)

# %%



import shlex
import subprocess
from unittest import TestCase

import pandas
from pandas.testing import assert_frame_equal
from toolbox_utils import tsutils

from tstoolbox import tstoolbox

from . import capture


class TestPick(TestCase):
    def setUp(self):
        dr = pandas.date_range("2000-01-01", periods=6, freq="D")

        ts1 = pandas.Series([4.5, 4.6, 4.7, 4.6, 4.5, 4.4], index=dr)
        ts2 = pandas.Series([45.6, 90.5, 34.2, 23.1, 7.2, 4.3], index=dr)
        self.pick_multiple_direct = pandas.DataFrame(ts2, columns=["Value1"])
        self.pick_multiple_direct = self.pick_multiple_direct.join(
            pandas.DataFrame(ts1, columns=["Value"])
        )
        self.pick_multiple_direct = tsutils.memory_optimize(self.pick_multiple_direct)

        self.pick_cli = capture.capture(tsutils.printiso, self.pick_multiple_direct)

    def test_pick(self):
        """Test the pick API by picking the 2nd then the 1st column.

        Reversing the order of the columns.
        """
        out = tstoolbox.pick("2,1", input_ts="tests/data_multiple_cols.csv")
        assert_frame_equal(out, self.pick_multiple_direct)

    def test_pick_cli(self):
        """Test the pick API by picking the 2nd then the 1st column.

        Effectively reversing the order of the columns.
        """
        args = 'tstoolbox pick 2,1 --input_ts="tests/data_multiple_cols.csv"'
        args = shlex.split(args)
        out = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()
        self.assertEqual(out[0], self.pick_cli)



"""Collection of functions for the manipulation of time series."""

import cltoolbox
from cltoolbox.rst_text_formatter import RSTHelpFormatter
from pydantic import validate_arguments
from toolbox_utils import tsutils


@cltoolbox.command("converttz", formatter_class=RSTHelpFormatter)
@tsutils.doc(tsutils.docstrings)
def converttz_cli(
    fromtz,
    totz,
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    round_index=None,
    dropna="no",
    clean=False,
    index_type="datetime",
    names=None,
    source_units=None,
    target_units=None,
    skiprows=None,
    tablefmt="csv",
):
    """Convert the time zone of the index.

    Parameters
    ----------
    fromtz : str
        The time zone of the original time-series.  The 'EST', and
        'America/New_York' could in some sense be thought of as the
        same, however 'EST' would force the time index to have the same
        offset from UTC, regardless of daylight savings time, where
        'America/New_York' would implement the appropriate daylight
        savings offsets.
    totz : str
        The time zone of the converted time-series.  Same note applies
        as for `fromtz`.  Needs to be different from `fromtz`.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${round_index}
    ${dropna}
    ${clean}
    ${skiprows}
    ${index_type}
    ${names}
    ${source_units}
    ${target_units}
    ${tablefmt}
    """
    tsutils.printiso(
        converttz(
            fromtz,
            totz,
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            round_index=round_index,
            dropna=dropna,
            clean=clean,
            index_type=index_type,
            names=names,
            source_units=source_units,
            target_units=target_units,
            skiprows=skiprows,
        ),
        showindex="always",
        tablefmt=tablefmt,
    )


@validate_arguments
@tsutils.copy_doc(converttz_cli)
def converttz(
    fromtz: str,
    totz: str,
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    round_index=None,
    dropna="no",
    clean=False,
    index_type="datetime",
    names=None,
    source_units=None,
    target_units=None,
    skiprows=None,
):
    """Convert the time zone of the index."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )

    # TODO Should test that 'fromtz' matches time zone that might be already
    # set in tsd.

    if totz == "None" or not totz:
        totz = None

    try:
        tsd = tsd.tz_localize(fromtz).tz_convert(totz)
    except TypeError:
        tsd = tsd.tz_convert(totz)
    tsd = tsutils.memory_optimize(tsd)
    return tsd



def report_duplicate_stats(duplicates):
    """
    Report list of pairs of statistics that are duplicates.

    INPUTS:
    DUPLICATES : List of tuples of paired statistics that are duplicates
                 produced by the check_duplicate_stats function

    OUTPUTS:
    None

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Apr 23, 2017
    """
    if len(duplicates) == 0:
        return

    # Report duplicates to screen
    print("Duplicate pairs of statistics:")
    for pair in duplicates:
        print(pair)



import shlex
import subprocess
from unittest import TestCase

import pandas as pd
from pandas.testing import assert_frame_equal
from toolbox_utils import tsutils

from tstoolbox import tstoolbox


class TestConvert(TestCase):
    def setUp(self):
        dr = pd.date_range("2000-01-01", periods=2, freq="D")
        ts = pd.Series([4.5, 4.6], index=dr)
        self.compare_direct_01 = pd.DataFrame(ts, columns=["Value::convert"])
        self.compare_direct_01.index.name = "Datetime"
        self.compare_direct_01 = tsutils.memory_optimize(self.compare_direct_01)

        dr = pd.date_range("2000-01-01", periods=2, freq="D")
        ts = pd.Series([11.0, 11.2], index=dr)
        self.compare_direct_02 = pd.DataFrame(ts, columns=["Value::convert"])
        self.compare_direct_02.index.name = "Datetime"
        self.compare_direct_02 = (
            tsutils.memory_optimize(self.compare_direct_02)
        ).astype("Float64")

        self.compare_cli_01 = b"""Datetime,Value::convert
2000-01-01,4.5
2000-01-02,4.6
"""
        self.compare_cli_02 = b"""Datetime,Value::convert
2000-01-01,11.9
2000-01-02,12.12
"""

    def test_convert_direct_01(self):
        """Test of convert API with default factor and offset."""
        out = tstoolbox.convert(input_ts="tests/data_simple.csv")
        assert_frame_equal(out, self.compare_direct_01)

    def test_convert_direct_02(self):
        """Test of convert API with set factor and offset."""
        out = (
            tstoolbox.convert(input_ts="tests/data_simple.csv", factor=2, offset=2)
        ).astype("Float64")
        assert_frame_equal(out, self.compare_direct_02)

    def test_convert_cli_01(self):
        """Test of CLI convert with default factor and offset."""
        args = 'tstoolbox convert --input_ts="tests/data_simple.csv"'
        args = shlex.split(args)
        out = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()
        self.assertEqual(out[0], self.compare_cli_01)

    def test_convert_cli_02(self):
        """Test of CLI convert set factor and offset."""
        args = (
            "tstoolbox convert "
            "--factor=2.2 "
            "--offset=2 "
            '--input_ts="tests/data_simple.csv"'
        )
        args = shlex.split(args)
        out = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()
        self.assertEqual(out[0], self.compare_cli_02)



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # tstoolbox: Statistics
# I will use the water level record for USGS site '02240000 OCKLAWAHA RIVER NEAR CONNER, FL'.  Use tsgettoolbox to download the data we will be using.

# %%
from tsgettoolbox import tsgettoolbox

from tstoolbox import tstoolbox

# %%
ock_flow = tsgettoolbox.nwis(
    sites="02240000", startDT="2008-01-01", endDT="2016-01-01", parameterCd="00060"
)

# %%
ock_flow.head()

# %%
ock_stage = tsgettoolbox.nwis(
    sites="02240000", startDT="2008-01-01", endDT="2016-01-01", parameterCd="00065"
)

# %%
ock_stage.head()

# %% [markdown]
# The tstoolbox.rolling_window calculates the average stage over the spans listed in the 'span' argument.

# %%
r_ock_stage = tstoolbox.rolling_window(
    input_ts=ock_stage, span="1,14,30,60,90,120,180,274,365", statistic="mean"
)
a_ock_stage = tstoolbox.aggregate(
    input_ts=r_ock_stage, agg_interval="A", statistic="max"
)

# %%
a_ock_stage

# %%
q_max_avg_obs = tstoolbox.calculate_fdc(input_ts=a_ock_stage, sort_order="descending")

# %%
q_max_avg_obs

# %%
tstoolbox.plot(
    input_ts=q_max_avg_obs,
    ofilename="q_max_avg.png",
    type="norm_xaxis",
    xtitle="Annual Exceedence Probability (%)",
    ytitle="Stage (feet)",
    title="N-day Maximum Rolling Window Average Stage for 0224000",
    legend_names="1,14,30,60,90,120,180,274,365",
)

# %% [markdown]
# ![](q_max_avg.png)

# %%
combined = ock_flow.join(ock_stage)
tstoolbox.plot(
    input_ts=combined,
    type="xy",
    ofilename="stage_flow.png",
    legend=False,
    title="Plot of Daily Stage vs Flow for 02240000",
    xtitle="Flow (cfs)",
    ytitle="Stage (ft)",
)

# %% [markdown]
# ![](stage_flow.png)

# %% [markdown]
# Note the signficant effect of tail-water elevation.

# %%



"""Collection of functions for the manipulation of time series."""


import warnings
from typing import Union

import cltoolbox
from cltoolbox.rst_text_formatter import RSTHelpFormatter
from pydantic import validate_arguments
from toolbox_utils import tsutils

warnings.filterwarnings("ignore")


@cltoolbox.command("unstack", formatter_class=RSTHelpFormatter)
@tsutils.doc(tsutils.docstrings)
def unstack_cli(
    column_names,
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    round_index=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    source_units=None,
    target_units=None,
    clean=False,
    tablefmt="csv",
):
    """Return the unstack of the input table.

    The unstack command takes the stacked table and converts to a
    standard tstoolbox table.

    From::

      Datetime,Columns,Values
      2000-01-01,TS1,1.2
      2000-01-02,TS1,1.8
      2000-01-03,TS1,1.9
      2000-01-01,TS2,1018.2
      2000-01-02,TS2,1453.1
      2000-01-03,TS2,1683.1
      2000-01-01,TS3,0.0032
      2000-01-02,TS3,0.0002
      2000-01-03,TS3,-0.0004

    To::

      Datetime,TS1,TS2,TS3
      2000-01-01,1.2,1018.2,0.0032
      2000-01-02,1.8,1453.1,0.0002
      2000-01-03,1.9,1683.1,-0.0004

    Parameters
    ----------
    column_names
        The column in the table that holds the column names
        of the unstacked data.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${skiprows}
    ${index_type}
    ${names}
    ${clean}
    ${source_units}
    ${target_units}
    ${round_index}
    ${tablefmt}
    """
    tsutils.printiso(
        unstack(
            column_names,
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            round_index=round_index,
            dropna=dropna,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            source_units=source_units,
            target_units=target_units,
            clean=clean,
        ),
        tablefmt=tablefmt,
    )


@validate_arguments
@tsutils.copy_doc(unstack_cli)
def unstack(
    column_names: Union[int, str],
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    round_index=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    source_units=None,
    target_units=None,
    clean=False,
):
    """Return the unstack of the input table."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        pick=columns,
        bestfreq=False,
        clean=False,
    )

    try:
        newtsd = tsd.pivot_table(
            index=tsd.index,
            values=tsd.columns.drop(column_names),
            columns=column_names,
            aggfunc="first",
        )
    except ValueError:
        raise ValueError(
            tsutils.error_wrapper(
                f"""
Duplicate index (time stamp and '{column_names}') where found.
Found these duplicate indices:
{tsd.index.get_duplicates()}
"""
            )
        )

    newtsd.index.name = "Datetime"

    newtsd.columns = [
        "_".join(tuple(map(str, col))).rstrip("_") for col in newtsd.columns.values
    ]

    # Remove weird characters from column names
    newtsd.rename(columns=lambda x: "".join([i for i in str(x) if i not in "'\" "]))

    newtsd = tsutils.common_kwds(
        newtsd,
        start_date=start_date,
        end_date=end_date,
        dropna=dropna,
        clean=False,
        source_units=source_units,
        target_units=target_units,
        round_index=round_index,
    )

    return newtsd



import numpy as np

from . import utils


def brier_score(forecast, observed):
    """
    Calculate Brier score (BS) between two variables.

    Calculates the Brier score (BS), a measure of the mean-square error of
    probability forecasts for a dichotomous (two-category) event, such as
    the occurrence/non-occurrence of precipitation. The score is calculated
    using the formula:

    BS = sum_(n=1)^N (f_n - o_n)^2/N

    where f is the forecast probabilities, o is the observed probabilities
    (0 or 1), and N is the total number of values in f & o. Note that f & o
    must have the same number of values, and those values must be in the
    range [0,1].

    Input:
    FORECAST : forecast probabilities
    OBSERVED : observed probabilities

    Output:
    BS : Brier score

    Reference:
    Glenn W. Brier, 1950: Verification of forecasts expressed in terms
    of probabilities. Mon. We. Rev., 78, 1-23.

    D. S. Wilks, 1995: Statistical Methods in the Atmospheric Sciences.
    Cambridge Press. 547 pp.

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Dec 15, 2016
    """
    # Check that dimensions of forecast and observed fields match
    utils.check_arrays(forecast, observed, pname="forecast", rname="observed")

    # Check for valid values
    index = np.where(np.logical_or(forecast < 0, forecast > 1))
    if np.sum(index) > 0:
        raise ValueError("Forecast has values outside interval [0,1].")
    index = np.where(np.logical_and(observed != 0, observed != 1))
    if np.sum(index) > 0:
        raise ValueError("Observed has values not equal to 0 or 1.")

    return np.sum(np.square(forecast - observed)) / len(forecast)



import shlex
import subprocess
import sys

from setuptools import setup

pkg_name = "tstoolbox"

with open("VERSION", encoding="ascii") as version_file:
    version = version_file.readline().strip()

if sys.argv[-1] == "publish":
    subprocess.run(shlex.split("cleanpy ."), check=True)
    subprocess.run(shlex.split("python setup.py sdist"), check=True)
    subprocess.run(
        shlex.split(f"twine upload dist/{pkg_name}-{version}.tar.gz"), check=True
    )
    sys.exit()

setup()



import matplotlib.pyplot as plt
from matplotlib import ticker


def plot_pattern_diagram_colorbar(X, Y, Z, option):
    """
    Plot color markers on a pattern diagram shaded to a supplied value.

    Values are indicated via a color bar on the plot.

    Plots color markers on a target diagram according their (X,Y) locations.
    The color shading is accomplished by plotting the markers as a scatter
    plot in (X,Y) with the colors of each point specified using Z as a
    vector.

    The color range is controlled by option['nonRMSDz'].
    option['colormap'] = 'on' :
        the scatter function maps the elements in Z to colors in the
        current colormap
    option['colormap']= 'off' : the color axis is mapped to the range
        [min(Z) max(Z)]

    The color bar is titled using the content of option['titleColorBar']
    (if non-empty string).

    INPUTS:
    x : x-coordinates of markers
    y : y-coordinates of markers
    z : z-coordinates of markers (used for color shading)
    option : dictionary containing option values.
    option['colormap'] : 'on'/'off' switch to map color shading of markers
        to colormap ('on') or min to max range of Z values ('off').
    option['titleColorBar'] : title for the color bar

    OUTPUTS:
    None.

    Created on Nov 30, 2016

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com
    """
    # Plot color shaded data points using scatter plot
    # Keyword s defines marker size in points^2
    #         c defines the sequence of numbers to be mapped to colors
    #           using the cmap and norm
    _ = plt.scatter(X, Y, s=50, c=Z, marker="d", edgecolors="none")

    # Add color bar to plot
    if option["colormap"] == "on":
        hc = plt.colorbar(orientation="horizontal", aspect=6, fraction=0.04, pad=0.04)

    elif option["colormap"] == "off":
        if len(Z) > 1:
            plt.clim(min(Z), max(Z))
            hc = plt.colorbar(
                orientation="horizontal", aspect=6, fraction=0.04, pad=0.04
            )
            hc.set_ticklabels("Min. RMSD", "Max. RMSD")
    else:
        raise ValueError(f"Invalid option for option.colormap: {option['colormap']}")

    # Set desired properties of color bar
    location = _getColorBarLocation(hc, option, xscale=0.75, yscale=7.5, cxscale=1.0)
    hc.ax.set_position(location)  # set new position
    hc.ax.tick_params(labelsize="medium")  # set tick labels to medium

    # Limit number of ticks on colar bar to 4
    hc.locator = ticker.MaxNLocator(nbins=5)
    hc.update_ticks()

    hc.ax.xaxis.set_ticks_position("top")
    hc.ax.xaxis.set_label_position("top")

    # Title the color bar
    if option["titlecolorbar"]:
        hc.set_label(option["titlecolorbar"])
    else:
        hc.set_label(hc, "Color Scale")


def _getColorBarLocation(hc, option, **kwargs):
    """
    Determine location for color bar.

    Determines location to place color bar for type of plot:
    target diagram and Taylor diagram. Optional scale arguments
    (xscale,cxscale) can be supplied to adjust the placement of
    the colorbar to accommodate different situations.

    INPUTS:
    hc     : handle returned by colorbar function
    option : dictionary containing option values. (Refer to
             display_target_diagram_options function for more
             information.)

    OUTPUTS:
    location : x, y, width, height for color bar

    KEYWORDS:
    xscale  : scale factor to adjust x-position of color bar
    yscale  : scale factor to adjust y-position of color bar
    cxscale : scale factor to adjust thickness of color bar
    """
    # Check for optional arguments and set defaults if required
    xscale = kwargs.get("xscale", 1.0)
    yscale = kwargs.get("yscale", 1.0)
    cxscale = kwargs.get("cxscale", 1.0)

    # Get current position of color bar
    cp = hc.ax.get_position()

    return (
        [
            cp.x0 + xscale * 0.8 * cp.width,
            yscale * cp.y0,
            cxscale * cp.width / 6,
            cp.height,
        ]
        if "checkSTATS" in option
        else [
            cp.x0 + xscale * cp.width,
            yscale * cp.y0,
            cxscale * cp.width / 6,
            cp.height,
        ]
    )



"""Collection of functions for the manipulation of time series."""

import warnings
from typing import Union

import cltoolbox
import pandas as pd
from cltoolbox.rst_text_formatter import RSTHelpFormatter
from pydantic import validate_arguments
from toolbox_utils import tsutils

warnings.filterwarnings("ignore")


@cltoolbox.command("createts", formatter_class=RSTHelpFormatter)
@tsutils.doc(tsutils.docstrings)
def createts_cli(
    freq=None,
    fillvalue=None,
    input_ts=None,
    index_type="datetime",
    start_date=None,
    end_date=None,
    tablefmt="csv",
):
    """Create empty time series, optionally fill with a value.

    Parameters
    ----------
    freq : str
        [optional, default is None]

        To use this form `start_date` and `end_date` must be supplied
        also.  The `freq` option is the pandas date offset code used to create
        the index.

        Python example::

            freq='A'

        Command line example::

            --freq='A'

        ${pandas_offset_codes}
    fillvalue
        [optional, default is None]

        The fill value for the time-series.  The default is None, which
        generates the date/time stamps only.
    ${input_ts}
    ${start_date}
    ${end_date}
    ${index_type}
    ${tablefmt}
    """
    tsutils.printiso(
        createts(
            freq=freq,
            fillvalue=fillvalue,
            input_ts=input_ts,
            index_type=index_type,
            start_date=start_date,
            end_date=end_date,
        ),
        showindex="always",
        tablefmt=tablefmt,
    )


@validate_arguments
@tsutils.copy_doc(createts_cli)
def createts(
    input_ts=None,
    freq: str = None,
    fillvalue: Union[float, int] = None,
    index_type="datetime",
    start_date=None,
    end_date=None,
):
    """Create empty time series, optionally fill with a value."""
    if input_ts is None and (
        (start_date is None) or (end_date is None) or (freq is None)
    ):
        raise ValueError(
            tsutils.error_wrapper(
                f"""
If input_ts is None, then start_date, end_date, and freq must be supplied.

Instead you have:
start_date = {start_date},
end_date = {end_date},
freq = {freq}
"""
            )
        )

    if input_ts is not None:
        tsd = tsutils.common_kwds(
            input_ts,
            index_type=index_type,
            start_date=start_date,
            end_date=end_date,
        )
        tsd = pd.DataFrame([fillvalue] * len(tsd.index), index=tsd.index)
    else:
        tindex = pd.date_range(start=start_date, end=end_date, freq=freq)
        tsd = pd.DataFrame([fillvalue] * len(tindex), index=tindex)
    return tsd



"""Collection of functions for the manipulation of time series."""

import warnings
from typing import List, Optional, Union

import cltoolbox
import pandas as pd
from cltoolbox.rst_text_formatter import RSTHelpFormatter
from pydantic import validate_arguments
from toolbox_utils import tsutils

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


warnings.filterwarnings("ignore")


@cltoolbox.command("accumulate", formatter_class=RSTHelpFormatter)
@tsutils.doc(tsutils.docstrings)
def accumulate_cli(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    clean=False,
    statistic="sum",
    round_index=None,
    skiprows=None,
    index_type="datetime",
    names=None,
    source_units=None,
    target_units=None,
    print_input=False,
    tablefmt="csv",
):
    """Calculate accumulating statistics.

    Parameters
    ----------
    statistic : Union(str, list(str))
        [optional, default is "sum", transformation]

        OneOrMore("sum", "max", "min", "prod")

        Python example::
            statistic=["sum", "max"]

        Command line example::
            --statistic=sum,max
    ${input_ts}
    ${start_date}
    ${end_date}
    ${skiprows}
    ${names}
    ${columns}
    ${dropna}
    ${clean}
    ${source_units}
    ${target_units}
    ${round_index}
    ${index_type}
    ${print_input}
    ${tablefmt}
    """
    tsutils.printiso(
        accumulate(
            input_ts=input_ts,
            skiprows=skiprows,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            dropna=dropna,
            clean=clean,
            statistic=statistic,
            round_index=round_index,
            index_type=index_type,
            names=names,
            source_units=source_units,
            target_units=target_units,
            print_input=print_input,
        ),
        tablefmt=tablefmt,
    )


@tsutils.transform_args(
    statistic=tsutils.make_list,
    columns=tsutils.make_list,
    names=tsutils.make_list,
    source_units=tsutils.make_list,
    target_units=tsutils.make_list,
)
@validate_arguments
@tsutils.copy_doc(accumulate_cli)
def accumulate(
    input_ts="-",
    columns: Optional[Union[str, List]] = None,
    start_date=None,
    end_date=None,
    dropna="no",
    clean=False,
    statistic: List[Literal["sum", "max", "min", "prod"]] = "sum",
    round_index=None,
    skiprows=None,
    index_type="datetime",
    names: Optional[List] = None,
    source_units: Optional[List] = None,
    target_units: Optional[List] = None,
    print_input=False,
):
    """Calculate accumulating statistics."""
    statistic = tsutils.make_list(statistic)
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )
    ntsd = pd.DataFrame()

    for stat in statistic:
        tmptsd = eval(f"tsd.cum{stat}()")
        tmptsd.columns = [tsutils.renamer(i, stat) for i in tmptsd.columns]
        ntsd = ntsd.join(tmptsd, how="outer")
    return tsutils.return_input(print_input, tsd, ntsd)



"""Collection of functions for the manipulation of time series."""

import warnings

import cltoolbox
from cltoolbox.rst_text_formatter import RSTHelpFormatter
from pydantic import validate_arguments
from toolbox_utils import tsutils

warnings.filterwarnings("ignore")


@cltoolbox.command("clip", formatter_class=RSTHelpFormatter)
@tsutils.doc(tsutils.docstrings)
def clip_cli(
    input_ts="-",
    start_date=None,
    end_date=None,
    columns=None,
    dropna="no",
    clean=False,
    skiprows=None,
    index_type="datetime",
    names=None,
    a_min=None,
    a_max=None,
    round_index=None,
    source_units=None,
    target_units=None,
    print_input=False,
    tablefmt="csv",
):
    """Return a time-series with values limited to [a_min, a_max].

    Parameters
    ----------
    a_min
        [optional, defaults to None, transformation]

        All values lower than this will be set to this value.
        Default is None.
    a_max
        [optional, defaults to None, transformation]

        All values higher than this will be set to this value.
        Default is None.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${clean}
    ${skiprows}
    ${index_type}
    ${print_input}
    ${names}
    ${source_units}
    ${target_units}
    ${round_index}
    ${tablefmt}
    """
    tsutils.printiso(
        clip(
            input_ts=input_ts,
            start_date=start_date,
            end_date=end_date,
            columns=columns,
            dropna=dropna,
            clean=clean,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            a_min=a_min,
            a_max=a_max,
            round_index=round_index,
            source_units=source_units,
            target_units=target_units,
            print_input=print_input,
        ),
        tablefmt=tablefmt,
    )


@validate_arguments
@tsutils.copy_doc(clip_cli)
def clip(
    input_ts="-",
    start_date=None,
    end_date=None,
    columns=None,
    dropna="no",
    clean=False,
    skiprows=None,
    index_type="datetime",
    names=None,
    a_min: float = None,
    a_max: float = None,
    round_index=None,
    source_units=None,
    target_units=None,
    print_input=False,
):
    """Return a time-series with values limited to [a_min, a_max]."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )
    ntsd = tsd.clip(lower=a_min, upper=a_max)

    return tsutils.return_input(print_input, tsd, ntsd, "clip")



"""Collection of functions for the manipulation of time series."""


import warnings
from typing import List, Optional, Union

import cltoolbox
from cltoolbox.rst_text_formatter import RSTHelpFormatter
from pydantic import validate_arguments
from toolbox_utils import tsutils

warnings.filterwarnings("ignore")


@cltoolbox.command("replace", formatter_class=RSTHelpFormatter)
@tsutils.doc(tsutils.docstrings)
def replace_cli(
    from_values,
    to_values,
    round_index=None,
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    source_units=None,
    target_units=None,
    print_input=False,
    tablefmt="csv",
):
    """Return a time-series replacing values with others.

    Parameters
    ----------
    from_values
        All values in this comma separated list are replaced with the
        corresponding value in to_values.  Use the string 'None' to
        represent a missing value.  If using 'None' as a from_value it
        might be easier to use the "fill" subcommand instead.
    to_values
        All values in this comma separated list are the replacement
        values corresponding one-to-one to the items in from_values.
        Use the string 'None' to represent a missing value.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${skiprows}
    ${index_type}
    ${names}
    ${clean}
    ${round_index}
    ${source_units}
    ${target_units}
    ${print_input}
    ${tablefmt}
    """
    tsutils.printiso(
        replace(
            from_values,
            to_values,
            round_index=round_index,
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            dropna=dropna,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            clean=clean,
            source_units=source_units,
            target_units=target_units,
            print_input=print_input,
        ),
        tablefmt=tablefmt,
    )


@tsutils.transform_args(from_values=tsutils.make_list, to_values=tsutils.make_list)
@validate_arguments
@tsutils.copy_doc(replace_cli)
def replace(
    from_values: Optional[List[Optional[Union[float, int, str]]]],
    to_values: Optional[List[Optional[Union[float, int, str]]]],
    round_index=None,
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    source_units=None,
    target_units=None,
    print_input=False,
):
    """Return a time-series replacing values with others."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )
    if from_values is None:
        from_values = [None]
    if to_values is None:
        to_values = [None]
    ntsd = tsd.replace(from_values, to_values)

    return tsutils.return_input(print_input, tsd, ntsd, "replace")



import numpy as np


def kge12(predicted, reference, sr=1.0, sgamma=1.0, sbeta=1.0):
    """
    Calculate the Kling-Gupta efficiency from 2012 paper.

    Calculates the Kling-Gupta efficiency between two variables
    predicted and reference. The kge12 is calculated using the
    formula:

    kge12 = 1 - sqrt((sr*(cc-1))**2 +
                     (sgamma*(gamma-1))**2 +
                     (sbeta*(beta-1))**2
                     )

    where:
        cc = correlation coefficient between predicted and reference;
        gamma = coefficient_of_variance(predicted) /
                coefficient_of_variance(reference)
        beta = sum(predicted) / sum(reference)

    where s is the predicted values, o is the reference values, and
    N is the total number of values in s & o. Note that s & o must
    have the same number of values.

    Kling-Gupta efficiency can range from -infinity to 1. An efficiency of 1 (E
    = 1) corresponds to a perfect match of predicted to reference data.

    The efficiency coefficient is sensitive to extreme values and might yield
    sub-optimal results when the dataset contains large outliers in it.

    Kling-Gupta efficiency can be used to quantitatively describe the accuracy
    of model outputs. This method can be used to describe the predictive
    accuracy of other models as long as there is reference data to compare the
    model results to.

    Input:
    predicted : predicted values
    reference : reference values
    sr : [optional, defaults to 1.0] scaling factor for correlation
    sgamma : [optional, defaults to 1.0] scaling factor for gamma
    sbeta : [optional, defaults to 1.0] scaling factor for beta

    Output:
    kge12 : Kling-Gupta Efficiency

    References:
    Kling, H., M. Fuchs, and M. Paulin (2012), Runoff conditions in the upper
    Danube basin under an ensemble of climate change scenarios. Journal of
    Hydrology, Volumes 424-425, 6 March 2012, Pages 264-277,
    DOI:10.1016/j.jhydrol.2012.01.011
    """
    # Check that dimensions of predicted and reference fields match
    pdims = predicted.shape
    rdims = reference.shape
    if not np.array_equal(pdims, rdims):
        message = f"predicted and reference field dimensions do not match.\nshape(predicted)= {str(pdims)}, shape(reference)= {str(rdims)}\npredicted type: {str(type(predicted))}"
        raise ValueError(message)

    for name, term in [("sr", sr), ("sgamma", sgamma), ("sbeta", sbeta)]:
        if term > 1 or term < 0:
            raise ValueError(f"'{name}' must be between 0 and 1, you gave {term}")

    std_ref = np.std(reference)
    if std_ref == 0:
        return -np.inf
    sum_ref = np.sum(reference)
    if sum_ref == 0:
        return -np.inf

    gamma = (np.std(predicted) / np.mean(predicted)) / (
        np.std(reference) / np.mean(reference)
    )
    beta = np.sum(predicted) / np.sum(reference)
    cc = np.corrcoef(reference, predicted)[0, 1]

    return 1.0 - np.sqrt(
        (sr * (cc - 1.0)) ** 2
        + (sgamma * (gamma - 1.0)) ** 2
        + (sbeta * (beta - 1.0)) ** 2
    )



import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt
import pytest
from toolbox_utils import tsutils

from tstoolbox import tstoolbox

# Pull this in once.
idf = tsutils.common_kwds(input_tsd="tests/02234500_65_65.csv", clean=True)

df = idf.resample("D").agg("mean")

dfa = idf.resample("A").agg("mean")


@pytest.mark.mpl_image_compare(tolerance=6)
def test_time_plot():
    plt.close("all")
    return tstoolbox.plot(
        type="time",
        columns=1,
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


# For some reason, these tests fail on Github workflows, but not locally.
# @pytest.mark.mpl_image_compare(tolerance=6)
# def test_time_multiple_traces_plot():
#    plt.close("all")
#    return tstoolbox.plot(
#        type="time",
#        columns=[2, 3],
#        style="b-,r*",
#        input_ts="tests/data_daily_sample.csv",
#        ofilename=None,
#        plot_styles="classic",
#    )
#
#
# @pytest.mark.mpl_image_compare(tolerance=6)
# def test_time_multiple_traces_style_plot():
#    plt.close("all")
#    return tstoolbox.plot(
#        type="time",
#        columns=[2, 3],
#        style="b-,r  ",
#        input_ts="tests/data_daily_sample.csv",
#        ofilename=None,
#        plot_styles="classic",
#    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_time_multiple_traces_new_style_plot():
    plt.close("all")
    return tstoolbox.plot(
        type="time",
        columns=[2, 3],
        markerstyles=" ,*",
        linestyles="-, ",
        input_ts="tests/data_daily_sample.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_time_markers():
    plt.close("all")
    return tstoolbox.plot(
        type="time",
        columns=[2, 3],
        linestyles=" ",
        markerstyles="auto",
        input_ts="tests/data_daily_sample.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_xy():
    plt.close("all")
    return tstoolbox.plot(
        type="xy",
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_xy_multiple_traces():
    plt.close("all")
    return tstoolbox.plot(
        type="xy",
        columns=[2, 3, 3, 2],
        input_ts="tests/data_daily_sample.csv",
        ofilename=None,
        plot_styles="classic",
    )


# @pytest.mark.mpl_image_compare(tolerance=6)
# def test_xy_multiple_traces_logy():
#    plt.close("all")
#    return tstoolbox.xy(
#        columns=[2, 3, 3, 2],
#        yaxis="log",
#        input_ts="tests/data_daily_sample.csv",
#        ofilename=None,
#        plot_styles="classic",
#    )


# @pytest.mark.mpl_image_compare(tolerance=6)
# def test_xy_multiple_traces_logx():
#    plt.close("all")
#    return tstoolbox.xy(
#        columns=[2, 3, 3, 2],
#        xaxis="log",
#        input_ts="tests/data_daily_sample.csv",
#        ofilename=None,
#        plot_styles="classic",
#    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_xy_multiple_traces_markers():
    plt.close("all")
    return tstoolbox.plot(
        type="xy",
        columns=[2, 3, 3, 2],
        linestyles=" ",
        markerstyles="auto",
        input_ts="tests/data_daily_sample.csv",
        ofilename=None,
        plot_styles="classic",
    )



"""
Capture function
----------------------------------

"""

import sys

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO


def capture(func, *args, **kwds):
    """Captures stdout from tests to test command line stdout."""
    sys.stdout = StringIO()  # capture output
    out = func(*args, **kwds)
    out = sys.stdout.getvalue()  # release output
    try:
        out = bytes(out, "utf8")
    except TypeError:
        pass
    return out



import shlex
import subprocess
from unittest import TestCase

import numpy as np
import pandas
from pandas.testing import assert_frame_equal

import tstoolbox.tsutils as tsutils
from tstoolbox import tstoolbox

from . import capture


class TestRollingWindow(TestCase):
    def setUp(self):
        dr = pandas.date_range("2000-01-01", periods=2, freq="D")
        ts = pandas.Series([np.nan, 9.1], index=dr)
        self.compare_rolling_window_sum = pandas.DataFrame(
            ts, columns=["Value::rolling.2.sum"]
        )
        self.compare_rolling_window_sum.index.name = "Datetime"

        dr = pandas.date_range("2000-01-01", periods=2, freq="D")
        ts = pandas.Series([np.nan, 4.55], index=dr)
        self.compare_rolling_window_mean = pandas.DataFrame(
            ts, columns=["Value::rolling.2.mean"]
        ).astype("Float64")
        self.compare_rolling_window_mean.index.name = "Datetime"

        self.compare_rolling_window_sum_cli = capture.capture(
            tsutils.printiso, self.compare_rolling_window_sum
        )

        self.compare_rolling_window_mean_cli = capture.capture(
            tsutils.printiso, self.compare_rolling_window_mean
        )

    def test_rolling_window_sum(self):
        """API: Rolling window sum for data_simple.csv is 9.1."""
        out = tstoolbox.rolling_window(
            statistic="sum", input_ts="tests/data_simple.csv"
        )
        assert_frame_equal(out, tsutils.read_iso_ts(self.compare_rolling_window_sum))

    def test_rolling_window_mean(self):
        """API: Rolling window mean for data_simple.csv is 4.55."""
        out = tstoolbox.rolling_window(
            statistic="mean", input_ts="tests/data_simple.csv"
        )
        assert_frame_equal(out, self.compare_rolling_window_mean)

    def test_rolling_window_sum_cli(self):
        """CLI: Rolling window mean for data_simple.csv is 9.1."""
        args = 'tstoolbox rolling_window sum --input_ts="tests/data_simple.csv"'
        args = shlex.split(args)
        out = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()[0]
        out = tsutils.read_iso_ts(out)
        assert_frame_equal(
            out, tsutils.read_iso_ts(self.compare_rolling_window_sum_cli)
        )

    def test_rolling_window_window_sum_cli(self):
        """CLI: Rolling window mean for data_simple.csv is 9.1."""
        args = (
            'tstoolbox rolling_window sum --window 2 --input_ts="tests/data_simple.csv"'
        )
        args = shlex.split(args)
        out = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()[0]
        out = tsutils.read_iso_ts(out)
        assert_frame_equal(
            out, tsutils.read_iso_ts(self.compare_rolling_window_sum_cli)
        )

    def test_rolling_window_mean_cli(self):
        """CLI: Rolling window sum for data_simple.csv is 4.55."""
        args = 'tstoolbox rolling_window mean --input_ts="tests/data_simple.csv"'
        args = shlex.split(args)
        out = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()[0]
        out = tsutils.read_iso_ts(out)
        assert_frame_equal(
            out, tsutils.read_iso_ts(self.compare_rolling_window_mean_cli)
        )
